import logging

from bluesky.preprocessors import finalize_wrapper

# from bluesky.suspenders import SuspendBoolLow
from bluesky.utils import Msg, make_decorator

from haven.instrument import beamline

log = logging.getLogger()


def shutter_suspend_wrapper(plan, shutter_signals=None):
    """
    Install suspenders to the RunEngine, and remove them at the end.

    Parameters
    ----------
    plan : iterable or iterator
        a generator, list, or similar containing `Msg` objects
    suspenders : suspender or list of suspenders
        Suspenders to use for the duration of the wrapper

    Yields
    ------
    msg : Msg
        messages from plan, with 'install_suspender' and 'remove_suspender'
        messages inserted and appended
    """
    if shutter_signals is None:
        shutters = beamline.devices.findall("shutters", allow_none=True)
        shutter_signals = [s.pss_state for s in shutters]
    # Create a suspender for each shutter
    suspenders = []

    ###################################################
    # Temporarily disabled for technical commissioning
    ###################################################
    # for sig in shutter_signals:
    #     suspender = SuspendBoolLow(sig, sleep=3.0)
    #     suspenders.append(suspender)
    # if not isinstance(suspenders, Iterable):
    #     suspenders = [suspenders]

    def _install():
        for susp in suspenders:
            yield Msg("install_suspender", None, susp)

    def _remove():
        for susp in suspenders:
            yield Msg("remove_suspender", None, susp)

    def _inner_plan():
        yield from _install()
        return (yield from plan)

    return (yield from finalize_wrapper(_inner_plan(), _remove()))


shutter_suspend_decorator = make_decorator(shutter_suspend_wrapper)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = "haven"
copyright = "2022, Argonne National Laboratory"
author = "Mark Wolfman"

# The full version, including alpha/beta/rc tags
release = "0.1.0"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.graphviz",
    "autoapi.sphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    # "nbsphinx",
]

autoapi_modules = {"haven": None}
autodoc_typehints = "description"
autodoc_mock_imports = [
    "mergedeep",
]
napoleon_numpy_docstring = True

# Add custom pre-amble to the rendered jupyter notebooks
# nbsphinx_prolog = r"""

# This file is also available as an interactive jupyter notebook.

# `Download this file as a notebook`_ (hint: right-click -> "save as").

# .. _Download this file as a notebook: {{ env.doc2path(env.docname, base=None)|string }}

# """

# nbsphinx_epilog = nbsphinx_prolog

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "classic"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = []



import logging
from typing import Sequence, Union  # , Iterable

from bluesky.preprocessors import baseline_wrapper as bluesky_baseline_wrapper
from bluesky.utils import make_decorator

from haven.instrument import beamline

log = logging.getLogger()


def baseline_wrapper(
    plan,
    devices: Union[Sequence, str] = [
        "motors",
        "power_supplies",
        "xray_sources",
        "APS",
        "baseline",
    ],
    name: str = "baseline",
):
    bluesky_baseline_wrapper.__doc__
    # Resolve devices
    devices = beamline.devices.findall(devices, allow_none=True)
    yield from bluesky_baseline_wrapper(plan=plan, devices=devices, name=name)


baseline_decorator = make_decorator(baseline_wrapper)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest

from firefly.run_browser.client import DatabaseWorker


@pytest.fixture()
async def worker(tiled_client):
    worker = DatabaseWorker(tiled_client)
    await worker.change_catalog("255id_testing")
    return worker


@pytest.mark.asyncio
async def test_catalog_names(worker):
    assert (await worker.catalog_names()) == ["255id_testing", "255bm_testing"]


@pytest.mark.asyncio
async def test_filter_runs(worker):
    runs = await worker.load_all_runs(filters={"plan": "xafs_scan"})
    # Check that the runs were filtered
    assert len(runs) == 1


@pytest.mark.asyncio
async def test_distinct_fields(worker):
    distinct_fields = await worker.load_distinct_fields()
    # Check that the dictionary has the right structure
    for key in ["sample_name"]:
        assert key in distinct_fields.keys()


async def test_stream_names(worker):
    uids = (await worker.catalog.client).keys()
    await worker.load_selected_runs(uids)
    stream_names = await worker.stream_names()
    assert stream_names == ["primary"]


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import uuid
from typing import Sequence

from bluesky import Msg
from bluesky import plan_stubs as bps
from ophyd import Device

from ..devices.shutter import ShutterState
from ..instrument import beamline
from ._shutters import close_shutters, open_shutters


def count_is_complete(*, old_value, value, **kwargs):
    """Check if the value is done."""
    was_running = old_value == 1
    is_running_now = value == 1
    is_done = was_running and not is_running_now
    return is_done


def record_dark_current(
    ion_chambers: Sequence[Device], shutters: Sequence[Device] = []
):
    """Record the dark current on the ion chambers.

    - Close shutters
    - Record ion chamber dark current
    - Restore shutters to their previous positions

    Parameters
    ==========
    ion_chambers
      Ion chamber devices or names.
    shutters
      Shutter devices or names. These shutters will be closed before
      recording the dark current, and then be returned to its original
      state afterward recording the dark current.

    """
    # Get previous shutter states
    old_shutters = {}
    for shutter in shutters:
        old_shutters[shutter] = yield from bps.rd(shutter.readback)
    # Close shutters
    yield from close_shutters(shutters)
    # Measure the dark current
    ion_chambers = beamline.devices.findall(ion_chambers)
    # Record dark currents
    group = uuid.uuid4()
    for ic in ion_chambers:
        yield Msg("trigger", ic, group=group, record_dark_current=True)
        # yield from bps.trigger(ic.record_dark_current, group=group, wait=False)
    # Wait for the devices to be done recording dark current
    yield from bps.wait(group=group)
    # Reset shutters to their original states
    to_open = [
        sht for sht, old_state in old_shutters.items() if old_state == ShutterState.OPEN
    ]
    yield from open_shutters(to_open)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest
from bluesky.plan_stubs import trigger_and_read
from bluesky.plans import count
from bluesky.protocols import Triggerable
from ophyd_async.core import Device

from haven import open_shutters_wrapper
from haven.preprocessors.open_shutters import open_fast_shutters_wrapper


@pytest.fixture()
def detector(sim_registry):
    class Detector(Triggerable, Device):
        """A stubbed detector that can be triggered."""

        def trigger(self):
            pass

    det = Detector(name="detector")
    return det


@pytest.fixture()
def fast_shutter(sim_registry):
    class FastShutter(Device):
        _ophyd_labels_ = {"shutters", "fast_shutters"}

    shutter = FastShutter(name="fast_shutter")
    sim_registry.register(shutter)
    return shutter


@pytest.fixture()
def bad_shutter(sim_registry):
    class BadShutter(Device):
        _ophyd_labels_ = {"shutters"}
        allow_close = False

    shutter = BadShutter(name="bad_shutter")
    sim_registry.register(shutter)
    return shutter


@pytest.fixture()
def slow_shutter(sim_registry):
    class SlowShutter(Device):
        _ophyd_labels_ = {"shutters"}

    shutter = SlowShutter(name="slow_shutter")
    sim_registry.register(shutter)
    return shutter


def test_slow_shutter_wrapper(
    sim_registry, detector, bad_shutter, slow_shutter, fast_shutter
):
    # Build the wrapped plan
    plan = trigger_and_read([detector])
    plan = open_shutters_wrapper(plan, registry=sim_registry)
    # Check that the current shutter position was read
    read_msg = next(plan)
    assert read_msg.command == "read"
    assert read_msg.obj in [slow_shutter, fast_shutter]
    read_msg = plan.send({read_msg.obj.name: {"value": 1}})
    assert read_msg.command == "read"
    assert read_msg.obj in [slow_shutter, fast_shutter]
    # Check that the shutter was opened
    set_msg = plan.send({read_msg.obj.name: {"value": 1}})
    assert set_msg.command == "set"
    assert set_msg.obj is slow_shutter
    assert set_msg.args[0] == 0
    wait_msg = next(plan)
    assert wait_msg.kwargs["group"] == set_msg.kwargs["group"]
    # Were the shutters closed again?
    msgs = list(plan)
    set_msg = msgs[-3]
    assert set_msg.command == "set"
    assert set_msg.obj in [slow_shutter, fast_shutter]
    assert set_msg.args[0] == 1
    set_msg = msgs[-2]
    assert set_msg.command == "set"
    assert set_msg.obj in [slow_shutter, fast_shutter]
    assert set_msg.args[0] == 1
    wait_msg = msgs[-1]
    assert wait_msg.kwargs["group"] == set_msg.kwargs["group"]


def test_fast_shutter_wrapper(sim_registry, detector, fast_shutter):
    # Build the wrapped plan
    plan = count([detector], num=2)
    plan = open_fast_shutters_wrapper(plan, shutters=[fast_shutter])
    msgs = list(plan)
    # Check that the shutter opens before triggering
    trig_idxs = [idx for idx, msg in enumerate(msgs) if msg.command == "trigger"]
    for trig_idx in trig_idxs:
        set_msg = msgs[trig_idx - 2]
        assert set_msg.command == "set"
        assert set_msg.obj == fast_shutter
        assert set_msg.args == (0,)
    # Check that the shutter closes after waiting
    for idx in trig_idxs:
        set_msg = msgs[idx + 2]
        assert set_msg.command == "set"
        assert set_msg.obj == fast_shutter
        assert set_msg.args == (1,)



import logging
from typing import Mapping, Optional

from bluesky_queueserver_api import comm_base
from bluesky_queueserver_api.zmq.aio import REManagerAPI
from qasync import asyncSlot
from qtpy.QtCore import QObject, QTimer, Signal

from haven import load_config
from haven.exceptions import InvalidConfiguration

log = logging.getLogger()


def is_in_use(status):
    # Add a new key for whether the queue is busy (length > 0 or running)
    has_queue = status.get("items_in_queue", 0) > 0
    is_running = status.get("manager_state") in [
        "paused",
        "starting_queue",
        "executing_queue",
        "executing_task",
    ]
    return has_queue or is_running


def queueserver_api():
    try:
        config = load_config()["queueserver"]
        ctrl_addr = f"tcp://{config['control_host']}:{config['control_port']}"
        info_addr = f"tcp://{config['info_host']}:{config['info_port']}"
    except KeyError as e:
        raise InvalidConfiguration(str(e))
    api = REManagerAPI(zmq_control_addr=ctrl_addr, zmq_info_addr=info_addr)
    return api


def queue_status(status_mapping: Mapping[str, str] = {}):
    """A generative coroutine that tracks the status of the queueserver.

    Parameters
    ==========
    status_mapping
      Maps the queuestatus parameters that are sent onto the update
      signal names to yield

    Yields
    ======
    to_update
      Dictionary with signals to emit as keys, and tuples of ``*args``
      to emit as values. So ``{self.status_changed: ("spam",
      "eggs")}`` results in ``self.status_changed.emit("spam",
      "eggs")``.

    Sends
    =====
    status
      The most recent updated status from the queueserver.

    """
    to_update = {}
    last_status = {}
    was_in_use = None
    while True:
        status = yield to_update
        to_update = {}
        if status != last_status:
            to_update["status_changed"] = (status,)
        # Check individual status items to see if they've changed
        status_diff = {
            key: val
            for key, val in status.items()
            if key not in last_status or val != last_status[key]
        }
        log.debug(f"Received updated queue status: {status_diff}")
        updated_params = {
            status_mapping[key]: (val,)
            for key, val in status_diff.items()
            if key in status_mapping
        }
        to_update.update(updated_params)
        # Decide if the queue is being used
        now_in_use = is_in_use(status)
        if now_in_use != was_in_use:
            to_update["in_use_changed"] = (now_in_use,)
            was_in_use = now_in_use
        # Stash this status for the next time around
        last_status = status


class QueueClient(QObject):
    api: REManagerAPI
    _last_queue_status: Optional[dict] = None
    last_update: float = -1
    timeout: float = 0.5
    timer: QTimer
    parameter_mapping: Mapping[str, str] = {
        "queue_autostart_enabled": "autostart_changed",
        "worker_environment_exists": "environment_opened",
        "worker_environment_state": "environment_state_changed",
        "manager_state": "manager_state_changed",
        "re_state": "re_state_changed",
        "devices_allowed_uid": "devices_allowed_changed",
    }

    # Signals responding to queue changes
    status_changed = Signal(dict)
    length_changed = Signal(int)
    in_use_changed = Signal(bool)  # If length > 0, or queue is running
    autostart_changed = Signal(bool)
    queue_stop_changed = Signal(bool)  # If a queue stop has been requested
    environment_opened = Signal(bool)  # Opened (True) or closed (False)
    environment_state_changed = Signal(str)  # New state
    manager_state_changed = Signal(str)  # New state
    re_state_changed = Signal(str)  # New state
    devices_changed = Signal(dict)

    def __init__(self, *args, api, **kwargs):
        self.api = api
        super().__init__(*args, **kwargs)
        self._last_queue_status = {}
        # Setup timer for updating the queue
        self.timer = QTimer()
        self.timer.timeout.connect(self.update)
        # Create the generator to keep track of the queue states
        self.status = queue_status(status_mapping=self.parameter_mapping)
        next(self.status)  # Prime the generator

    def start(self):
        # Start the time so that it triggers status updates
        self.timer.start(int(self.timeout * 1000))

    @asyncSlot(bool)
    async def open_environment(self, to_open):
        if to_open:
            result = await self.api.environment_open()
        else:
            result = await self.api.environment_close()
        if result["success"]:
            self.environment_opened.emit(to_open)
        else:
            log.error(f"Failed to open/close environment: {result['msg']}")

    @asyncSlot(bool)
    async def request_pause(self, *args, defer: bool = True, **kwargs):
        """Ask the queueserver run engine to pause.

        Parameters
        ==========
        defer
          If true, the run engine will be paused at the next
          checkpoint. Otherwise, it will pause now.

        """
        option = "deferred" if defer else "immediate"
        await self.api.re_pause(option=option)

    @asyncSlot(object)
    async def add_queue_item(self, item):
        log.info(f"Client adding item to queue: {item}")
        try:
            result = await self.api.item_add(item=item)
            self.check_result(result)
        except (RuntimeError, comm_base.RequestFailedError) as ex:
            # Request failed, so force a UI update
            raise
        finally:
            await self.update()

    @asyncSlot(bool)
    async def toggle_autostart(self, enable: bool):
        log.debug(f"Toggling auto-start: {enable}")
        try:
            result = await self.api.queue_autostart(enable)
            self.check_result(result, task="toggle auto-start")
        finally:
            await self.update()

    @asyncSlot(bool)
    async def stop_queue(self, stop: bool):
        """Turn on/off whether the queue will stop after the current plan."""
        # Determine which call to usee
        if stop:
            api_call = self.api.queue_stop()
        else:
            api_call = self.api.queue_stop_cancel()
        # Execute the call
        try:
            result = await api_call
            self.check_result(result, task="toggle stop queue")
        finally:
            await self.update()

    @asyncSlot()
    async def start_queue(self):
        result = await self.api.queue_start()
        self.check_result(result, task="start queue")

    @asyncSlot()
    async def resume_runengine(self):
        result = await self.api.re_resume()
        self.check_result(result, task="resume run engine")

    @asyncSlot()
    async def stop_runengine(self):
        result = await self.api.re_stop()
        self.check_result(result, task="stop run engine")

    @asyncSlot()
    async def abort_runengine(self):
        result = await self.api.re_abort()
        self.check_result(result, task="abort run engine")

    @asyncSlot()
    async def halt_runengine(self):
        result = await self.api.re_halt()
        self.check_result(result, task="halt run engine")

    def check_result(self, result: Mapping, task: str = "control queue server"):
        """Send the result of an API call to the correct logger.

        Expects *result* to have at least the "success" key.

        """
        # Report results
        if result["success"] is True:
            log.debug(f"{task}: {result}")
        else:
            msg = f"Failed to {task}: {result}"
            log.error(msg)
            raise RuntimeError(msg)

    @asyncSlot()
    async def update(self):
        """Get an update queue status from queue server and notify slots.

        Emits
        =====
        self.status_changed
          With the updated queue status.

        """
        new_status = await self.queue_status()
        signals_changed = self.status.send(new_status)
        # Check individual components of the status if they've changed
        if signals_changed != {}:
            log.debug(f"Emitting changed signals: {signals_changed}")
        for signal_name, args in signals_changed.items():
            if hasattr(self, signal_name):
                signal = getattr(self, signal_name)
                signal.emit(*args)
        # Check for new available devices
        if "devices_allowed_changed" in signals_changed:
            await self.update_devices()

    async def queue_status(self) -> dict:
        """Get the latest queue status from the queue server.

        Parameters
        ==========
        status
          The response from the queueserver regarding its status. If
          the queueserver is not reachable, then
          ``status['manager_state']`` will be ``"disconnected"``.

        """

        try:
            status = await self.api.status()
        except comm_base.RequestTimeoutError as e:
            log.warning("Could not reach queueserver ZMQ.")
            status = {
                "manager_state": "N.C.",
                "worker_environment_state": "N.C.",
                "re_state": "N.C.",
            }
        return status

    async def update_devices(self):
        "Emit the latest dict of available devices."
        response = await self.api.devices_allowed()
        if response["success"]:
            devices = response["devices_allowed"]
            self.devices_changed.emit(devices)
        else:
            log.warning(
                "Could not poll devices_allowed:"
                f" {response.get('msg', 'reason unknown.')}"
            )


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
import sys
from collections import defaultdict
from contextlib import contextmanager
from enum import Enum, IntEnum
from functools import partial
from itertools import product
from pathlib import Path
from typing import Sequence

import numpy as np
import pandas as pd
import pydm
import pyqtgraph
import qtawesome as qta
from matplotlib.colors import TABLEAU_COLORS
from qasync import asyncSlot
from qtpy import uic
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import QApplication, QLabel, QWidget

from firefly import display
from haven import beamline

np.set_printoptions(threshold=sys.maxsize)


log = logging.getLogger(__name__)


pyqtgraph.setConfigOption("imageAxisOrder", "row-major")

colors = list(TABLEAU_COLORS.values())


class AcquireStates(IntEnum):
    DONE = 0
    ACQUIRING = 1


class XRF1DPlotItem(pyqtgraph.PlotItem):
    """The axes used for plotting."""

    hover_coords_changed = Signal(str)

    def hoverEvent(self, event):
        super().hoverEvent(event)
        if event.isExit():
            self.hover_coords_changed.emit("NaN")
            return
        # Get data coordinates from event
        pos = event.scenePos()
        data_pos = self.vb.mapSceneToView(pos)
        pos_str = f"({data_pos.x():.3f}, {data_pos.y():.3f})"
        self.hover_coords_changed.emit(pos_str)


class XRF1DPlotWidget(pyqtgraph.PlotWidget):
    """The inner widget containing just the plot."""

    def __init__(self, parent=None, background="default", plotItem=None, **kargs):
        plot_item = XRF1DPlotItem(**kargs)
        plot_item.getAxis("bottom").setLabel("Energy", units="eV")
        super().__init__(parent=parent, background=background, plotItem=plot_item)


class XRFPlotWidget(QWidget):
    """The outer widget, containing the plot and related controls."""

    ui_dir = Path(__file__).parent
    _data_items: defaultdict
    _selected_spectrum: int | None = None
    device_name: str = ""
    target_mca: int = None

    # Signals
    plot_changed = Signal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._data_items = defaultdict(lambda: None)
        self.ui = uic.loadUi(self.ui_dir / "xrf_plot.ui", self)
        # Create plotting items
        plot_item = self.ui.plot_widget.getPlotItem()
        plot_item.addLegend()
        plot_item.hover_coords_changed.connect(self.ui.coords_label.setText)

    def update_spectrum(self, mca_num: int, spectrum: pd.Series):
        """Plot the spectrum associated with the given MCA index.

        Parameters
        ==========
        mca_num
          The 0-index for this MCA. E.g. the 2nd element would have
          *mca_num* of ``1``.
        spectrum
          The spectrum to plot.
        """
        # Create the plot item itself if necessary
        log.debug(f"New spectrum ({mca_num}): {repr(spectrum)}")
        show_spectrum = self.target_mca is None or mca_num == self.target_mca
        row, col = (0, 0)
        plot_item = self.ui.plot_widget.getPlotItem()
        # Get rid of the previous plots
        if (existing_item := self._data_items[mca_num]) is not None:
            plot_item.removeItem(existing_item)
        # Plot the spectrum
        if show_spectrum:
            try:
                length = len(spectrum)
            except TypeError:
                # Probably this means the spectrum is really just a scaler
                length = 1
                spectrum = np.asarray([spectrum])
            xdata = spectrum.index
            color = self.spectrum_color(mca_num)
            self._data_items[mca_num] = plot_item.plot(
                xdata, spectrum.values, name=mca_num, pen=color
            )
            # Add region markers
            self.plot_changed.emit()

    def spectrum_color(self, mca_num):
        return colors[(mca_num) % len(colors)]

    def highlight_spectrum(self, mca_num, roi_num, hovered):
        """Highlight a spectrum and lowlight the rest.

        *mca_num* is the number (1-indexed) of the spectrum plotted.

        If *hovered* is true, then set the highlights, otherwise clear
        them.

        """
        raise NotImplementedError


class MCAPlotWidget(XRFPlotWidget):
    def region_color(self, mca_num, roi_num):
        return self.spectrum_color(mca_num)

    def highlight_spectrum(self, mca_num, roi_num, hovered):
        """Highlight a spectrum and lowlight the rest.

        *mca_num* is the number (1-indexed) of the spectrum plotted.

        If *hovered* is true, then set the highlights, otherwise clear
        them.

        """
        # Get the actual line plot
        solid, semisolid, transparent = (1.0, 0.55, 0.15)
        for key, data_item in self._data_items.items():
            if data_item is None:
                continue
            elif hovered:
                # Highlight this specific spectrum
                if key == mca_num:
                    opacity = solid
                elif key == self._selected_spectrum:
                    opacity = semisolid
                else:
                    opacity = transparent
                data_item.setOpacity(opacity)
            elif self._selected_spectrum is not None:
                # Highlight the spectrum that was previously selected
                log.debug(
                    "Reverting to previously selected spectrum:"
                    f" {self._selected_spectrum}"
                )
                is_dimmed = key != self._selected_spectrum
                if is_dimmed:
                    data_item.setOpacity(transparent)
                else:
                    data_item.setOpacity(solid)
            else:
                # Just make them all solid again
                data_item.setOpacity(solid)

    def select_spectrum(self, mca_num, roi_num, is_selected):
        """Select an active spectrum to modify."""
        log.debug(f"Selecting spectrum {mca_num}")
        self._selected_spectrum = mca_num if is_selected else None
        self.highlight_spectrum(mca_num=mca_num, roi_num=roi_num, hovered=is_selected)


class ROIPlotWidget(XRFPlotWidget):
    def region_color(self, mca_num, roi_num):
        return colors[(roi_num) % len(colors)]

    def select_roi(self, mca_num, roi_num, is_selected):
        """Select an active spectrum to modify."""
        log.debug(f"Selecting ROI {roi_num}")
        self._selected_roi = roi_num if is_selected else None
        self.highlight_spectrum(mca_num=mca_num, roi_num=roi_num, hovered=is_selected)

    def highlight_spectrum(self, mca_num, roi_num, hovered):
        """Highlight a spectrum and lowlight the rest.

        *mca_num* is the number (1-indexed) of the spectrum plotted.

        If *hovered* is true, then set the highlights, otherwise clear
        them.

        """
        # Hide or unhide the selecting region
        is_selected = self._selected_roi is not None
        if hovered:
            self.show_region(show=True, mca_num=mca_num, roi_num=roi_num)
        elif is_selected:
            self.show_region(show=True, mca_num=mca_num, roi_num=self._selected_roi)
        else:
            self.show_region(show=False, mca_num=mca_num, roi_num=roi_num)


class Color(str, Enum):
    """Taken from bootstrap 5 alert components."""

    BLUE = "rgb(5, 81, 96)"
    GREY = "rgb(226, 227, 229)"
    GREEN = "rgb(10, 54, 34)"
    RED = "rgb(248, 215, 218)"
    YELLOW = "rgb(255, 243, 205)"


class XRFDetectorDisplay(display.FireflyDisplay):

    _spectrum_channels: Sequence
    _spectra: dict
    num_header_rows: int = 2

    # For styling the detector state attribute
    state_styles = {
        "Acquire": f"color: {Color.GREEN}; font-weight: bold;",
        "Saving": f"color: {Color.GREEN}",
        "Error": f"color: {Color.RED}; font-weight: bold;",
        "Disconnected": f"color: {Color.RED}",
        "Aborting": f"color: {Color.YELLOW}",
        "Initializing": f"color: {Color.YELLOW}",
        "Waiting": f"color: {Color.YELLOW}",
    }

    def ui_filename(self):
        return "xrf_detector.ui"

    def customize_ui(self):
        self._spectra = {}
        device = self.device
        self.setWindowTitle(device.name)
        self.ui.mca_plot_widget.device_name = self.device.name
        # Label for the device_name
        self.ui.detector_name_label.setText(device.name)
        # Create count totals for each element
        self.draw_mca_widgets()
        # Button for starting/stopping the detector
        self.ui.acquire_button.setIcon(qta.icon("fa5s.camera"))
        # Handler for updating the detector state label style
        self.det_state_channel = pydm.PyDMChannel(
            address=self.ui.detector_state_label.channel,
            value_slot=self.update_state_style,
        )
        self.det_state_channel.connect()
        super().customize_ui()

    def update_state_style(self, new_state: str):
        new_style = self.state_styles.get(new_state, "")
        self.ui.detector_state_label.setStyleSheet(new_style)

    @asyncSlot(object)
    async def handle_new_spectrum(self, new_spectrum, mca_num):
        # Calclulate energies for this spectrum
        ev_per_bin = await self.device.ev_per_bin.get_value()
        energies = (np.arange(new_spectrum.shape[0]) + 0.5) * ev_per_bin
        spectrum = pd.Series(new_spectrum, index=energies)
        self._spectra[mca_num] = spectrum
        # Update UI widgets
        self.ui.mca_plot_widget.update_spectrum(mca_num=mca_num, spectrum=spectrum)
        self.update_spectral_widgets(
            mca_num=mca_num, spectrum=spectrum, spectra=self._spectra.values()
        )

    def update_spectral_widgets(
        self, mca_num: int, spectrum: Sequence, spectra: Sequence[np.ndarray]
    ):
        """Update the values of labels with totals, etc from the spectra."""
        row = mca_num + self.num_header_rows
        count_label = self.ui.mcas_layout.itemAtPosition(row, 1).widget()
        count_label.setText(f"{np.sum(spectrum):_}")
        # Update the sum-total widget
        total = 0
        total_label = self.ui.mcas_layout.itemAtPosition(1, 1).widget()
        for spec in spectra:
            total += np.sum(spec)
        total_label.setText(f"{total:_}")

    def customize_device(self):
        # Load the device from the registry
        device_name = self.macros()["DEV"]
        self.device = beamline.devices[device_name]
        # Set up data channels
        self._spectrum_channels = []
        for element_num, element in self.device.elements.items():
            address = f"haven://{element.name}.spectrum"
            channel = pydm.PyDMChannel(
                address=address,
                value_slot=partial(self.handle_new_spectrum, mca_num=element_num),
            )
            channel.connect()
            self._spectrum_channels.append(channel)

    @contextmanager
    def disable_ui(self):
        widget = self
        # Disable to widgets
        widget.setEnabled(False)
        # Set waiting cursor
        old_cursor = self.cursor()
        self.setCursor(Qt.WaitCursor)
        # Update the UI
        QApplication.instance().processEvents()
        yield
        # Re-enabled everything
        widget.setEnabled(True)
        widget.setCursor(old_cursor)

    def draw_mca_widgets(self):
        """Prepare a row for each element in the detector."""
        # Prepare all the ROI widgets
        layout = self.ui.mcas_layout
        self.clear_elements_layout(layout)
        self._count_labels = {}
        for idx, (key, mca) in enumerate(self.device.elements.items()):
            row = idx + self.num_header_rows
            # Label for the number of the MCA's detector element
            key_label = QLabel()
            key_label.setText(str(key))
            layout.addWidget(key_label, row, 0)
            # Label for the counts for this element
            count_label = QLabel()
            count_label.setText("##########")
            layout.addWidget(count_label, row, 1)
            self._count_labels[key] = count_label
            # Label for the dead time for this element
            dt_signal = mca.dead_time_percent.name
            dt_label = pydm.widgets.PyDMLabel(init_channel=f"haven://{dt_signal}")
            layout.addWidget(dt_label, row, 2)

    def clear_elements_layout(self, layout):
        rows = range(self.num_header_rows, layout.rowCount())
        cols = range(layout.columnCount())
        for row, col in product(rows, cols):
            item = layout.itemAtPosition(row, col)
            layout.removeItem(item)
            item.widget().deleteLater()


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from bluesky import plan_stubs as bps
from bluesky.preprocessors import finalize_wrapper
from bluesky.utils import make_decorator
from ophydregistry import Registry

from haven.devices.shutter import ShutterState
from haven.instrument import beamline

__all__ = ["open_shutters_wrapper", "open_shutters_decorator"]


def _can_open(shutter):
    return getattr(shutter, "allow_open", True) and getattr(
        shutter, "allow_close", True
    )


def _set_shutters(shutters, state: int):
    """A plan stub that sets all shutters to the given state."""
    mv_args = [val for shutter in shutters for val in (shutter, state)]
    if len(mv_args) > 0:
        return (yield from bps.mv(*mv_args))
    else:
        yield from bps.null()


def open_shutters_wrapper(plan, registry: Registry | None = None):
    """Wrapper for Bluesky plans that opens and closes shutters as needed.

    Only shutters that are closed at the start of the plan are
    included.

    Shutters are split into two categories. **Fast shutters** (with
    the ophyd label ``"fast_shutters"``) will be **opened before a new
    detector trigger**, and closed after the trigger is awaited. All
    other shutters will be **opened at the start of the run**. Both
    categories will be closed at the end of the run.

    Parameters
    ==========
    plan
      The Bluesky plan instance to decorate.
    registry
      An ophyd-registry in which to look for shutters.

    """
    if registry is None:
        registry = beamline.devices
    # Get a list of shutters that could be opened and closed
    all_shutters = registry.findall(label="shutters", allow_none=True)
    all_shutters = [shtr for shtr in all_shutters if _can_open(shtr)]
    # Check for closed shutters (open shutters just stay open)
    shutters_to_open = []
    for shutter in all_shutters:
        initial_state = yield from bps.rd(shutter)
        if initial_state == ShutterState.CLOSED:
            shutters_to_open.append(shutter)
    # Organize the shutters into fast and slow
    fast_shutters = registry.findall(label="fast_shutters", allow_none=True)
    fast_shutters = [shtr for shtr in shutters_to_open if shtr in fast_shutters]
    slow_shutters = [shtr for shtr in shutters_to_open if shtr not in fast_shutters]
    # Open shutters
    yield from _set_shutters(slow_shutters, ShutterState.OPEN)
    # Add the wrapper for opening fast shutters at every trigger
    new_plan = open_fast_shutters_wrapper(plan, fast_shutters)
    # Add a wrapper to close all the shutters once the measurement is done
    close_shutters = _set_shutters(shutters_to_open, ShutterState.CLOSED)
    new_plan = finalize_wrapper(new_plan, close_shutters)
    # Execute the wrapped plan
    return_val = yield from new_plan
    return return_val


def open_fast_shutters_wrapper(plan, shutters):
    """Open and close each shutter when encountering a detector trigger."""
    response = None
    open_groups = set()
    while True:
        # Loop through the messages and get the next one in the queue
        try:
            msg = plan.send(response)
        except StopIteration as exc:
            return_val = exc.value
            break
        # Check for "trigger" messages to open shutters
        if msg.command == "trigger":
            if len(open_groups) == 0:
                yield from _set_shutters(shutters, ShutterState.OPEN)
            open_groups.add(msg.kwargs.get("group", None))
        # Emit the actual intended message
        response = yield msg
        # Check for "wait" messages to close shutters
        if msg.command == "wait":
            try:
                open_groups.remove(msg.kwargs.get("group", None))
            except KeyError:
                # Probably waiting on a group with no triggers
                pass
            else:
                if len(open_groups) == 0:
                    yield from _set_shutters(shutters, ShutterState.CLOSED)
    return return_val


open_shutters_decorator = make_decorator(open_shutters_wrapper)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest
from ophyd.utils.errors import ReadOnlyError
from ophyd_async.testing import get_mock_put, set_mock_value

from haven.devices.shutter import PssShutter, ShutterState


@pytest.fixture()
async def shutter(sim_registry):
    """
    Example PVs:

    S25ID-PSS:SCS:OpenEPICSC
    S25ID-PSS:SCS:CloseEPICSC
    S25ID-PSS:SCS:BeamBlockingM.VAL
    """
    shutter = PssShutter(prefix="S255ID-PSS:SCS:", name="shutter")
    await shutter.connect(mock=True)
    return shutter


async def test_read_shutter(shutter):
    """The current state of the shutter should be readable.

    This makes it compatible with the ``open_shutters_wrapper``.

    """
    reading = await shutter.read()
    assert shutter.name in reading


async def test_shutter_setpoint(shutter):
    """When we open and close the shutter, do the right EPICS signals get
    set?

    """
    # Prepare some mocking so we can operate properly
    open_put = get_mock_put(shutter.open_signal)
    close_put = get_mock_put(shutter.close_signal)
    set_mock_value(shutter.open_signal, 0)
    set_mock_value(shutter.close_signal, 0)
    # Close the shutter
    set_mock_value(shutter.open_signal, 0)
    set_mock_value(shutter.close_signal, 0)
    status = shutter.set(ShutterState.CLOSED)
    set_mock_value(shutter.readback, ShutterState.CLOSED)
    await status
    assert not open_put.called
    close_put.assert_called_once_with(1, wait=False)
    # Open the shutter
    open_put.reset_mock()
    close_put.reset_mock()
    set_mock_value(shutter.close_signal, 0)
    set_mock_value(shutter.open_signal, 0)
    status = shutter.set(ShutterState.OPEN)
    set_mock_value(shutter.readback, ShutterState.OPEN)
    await status
    assert not close_put.called
    open_put.assert_called_once_with(1, wait=False)


async def test_shutter_check_value(shutter):
    # Check for non-sense values
    with pytest.raises(ValueError):
        await shutter.set(ShutterState.FAULT)
    # Test shutter allow_close
    shutter.allow_close = False
    with pytest.raises(ReadOnlyError):
        await shutter.set(ShutterState.CLOSED)
    # Test shutter allow_open
    shutter.allow_open = False
    with pytest.raises(ReadOnlyError):
        await shutter.set(ShutterState.OPEN)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



"""Bluesky plans specific to spectroscopy.

Includes some standard bluesky plans with decorators.

"""

import bluesky.plans as bp

from haven.instrument import beamline
from haven.preprocessors import (
    baseline_decorator,
    open_shutters_decorator,
    shutter_suspend_decorator,
)

from ._align_motor import align_motor
from ._auto_gain import auto_gain
from ._energy_scan import energy_scan
from ._fly import fly_scan, grid_fly_scan
from ._record_dark_current import record_dark_current  # noqa: F401
from ._robot_transfer_sample import robot_transfer_sample  # noqa: F401
from ._set_energy import set_energy  # noqa: F401
from ._shutters import close_shutters, open_shutters  # noqa: F401
from ._xafs_scan import xafs_scan


def chain(*decorators):
    """Chain several decorators together into one decorator.

    Will be applied in reverse order, so the first item in *decorators* will
    be the outermost decorator.

    """

    def decorator(f):
        for d in reversed(decorators):
            f = d(f)
        return f

    return decorator


all_decorators = chain(
    shutter_suspend_decorator(), open_shutters_decorator(), baseline_decorator()
)

# Apply decorators to Haven plans
align_motor = all_decorators(align_motor)
auto_gain = all_decorators(auto_gain)
energy_scan = all_decorators(energy_scan)
fly_scan = baseline_decorator()(fly_scan)
grid_fly_scan = baseline_decorator()(grid_fly_scan)
xafs_scan = all_decorators(xafs_scan)

# Apply all_decorators to standard Bluesky plans
count = all_decorators(bp.count)
grid_scan = all_decorators(bp.grid_scan)
list_scan = all_decorators(bp.list_scan)
rel_grid_scan = all_decorators(bp.rel_grid_scan)
rel_list_scan = all_decorators(bp.rel_list_scan)
rel_scan = all_decorators(bp.rel_scan)
scan = all_decorators(bp.scan)
scan_nd = all_decorators(bp.scan_nd)

# Remove foreign imports
del beamline
del open_shutters_decorator
del baseline_decorator
del shutter_suspend_decorator
del bp

# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
import sys
from typing import Mapping, Sequence

import msgpack
from bluesky.callbacks.tiled_writer import TiledWriter
from bluesky_kafka import BlueskyConsumer
from tiled.client import from_uri
from tiled.client.base import BaseClient

import haven

log = logging.getLogger(__name__)


class TiledConsumer(BlueskyConsumer):
    """Send Bluesky documents received from a Kafka broker to Tiled for writing.

    There is no default configuration. A reasonable configuration for production is
        consumer_config={
            "auto.offset.reset": "latest"
        }

    Parameters
    ----------
    tiled_client
      The top-level Tiled client object to use for writing.
    topic_catalog_map
      Translates Kafka topic names to Tiled catalogs. Each value
      should be the name of a catalog available directly under
      *tiled_client*.
    bootstrap_servers
        Kafka server addresses as strings such as
        ``["broker1:9092", "broker2:9092", "127.0.0.1:9092"]``
    group_id : str
        Required string identifier for the consumer's Kafka Consumer group.
    consumer_config : dict
        Override default configuration or specify additional configuration
        options to confluent_kafka.Consumer.
    polling_duration : float
        Time in seconds to wait for a message before running function work_during_wait
        in the _poll method. Default is 0.05.
    deserializer : function, optional
        Function to deserialize data. Default is msgpack.loads.

    """

    def __init__(
        self,
        tiled_client: BaseClient,
        topic_catalog_map: Mapping,
        bootstrap_servers: Sequence[str],
        group_id: str,
        consumer_config: Mapping | None = None,
        polling_duration: float | int = 0.05,
        deserializer=msgpack.loads,
    ):
        self.topic_catalog_map = topic_catalog_map
        # Create writers for each Tiled catalog
        catalog_names = set(topic_catalog_map.values())
        self.writers = {
            catalog: TiledWriter(tiled_client[catalog]) for catalog in catalog_names
        }
        super().__init__(
            topics=list(topic_catalog_map.keys()),
            bootstrap_servers=",".join(bootstrap_servers),
            group_id=group_id,
            consumer_config=consumer_config,
            polling_duration=polling_duration,
            deserializer=deserializer,
        )

    def process_document(self, topic: str, name: str, doc: Mapping) -> bool:
        """Write the Bluesky document to the associated Tiled catalog.

        Parameters
        ----------
        topic
            the Kafka topic of the message containing name and doc
        name
            bluesky document name: `start`, `descriptor`, `event`, etc.
        doc
            bluesky document

        Returns
        -------
        continue_polling
            return False to break out of the polling loop, return True to continue polling
        """
        catalog = self.topic_catalog_map[topic]
        log.info(f"Writing {name} doc from {topic=} to {catalog=}.")
        writer = self.writers[catalog]
        writer(name, doc)


def main():
    """Launch the tiled consumer."""
    logging.basicConfig(level=logging.INFO)
    config = haven.load_config()
    bootstrap_servers = ["localhost:9092"]
    topic_catalog_map = {
        "bluesky.documents.haven": "haven",
        "bluesky.documents.haven-dev": "haven-dev",
    }
    # Create a tiled writer that will write documents to tiled
    tiled_uri = config["database"]["tiled"]["uri"]
    tiled_api_key = config["database"]["tiled"]["api_key"]
    client = from_uri(tiled_uri, api_key=tiled_api_key, include_data_sources=True)

    # Create a Tiled consumer that will listen for new documents.
    consumer = TiledConsumer(
        tiled_client=client,
        topic_catalog_map=topic_catalog_map,
        bootstrap_servers=bootstrap_servers,
        group_id="tiled_writer",
        consumer_config={"auto.offset.reset": "latest"},
        polling_duration=1.0,
    )
    log.info("Starting Tiled consumer")
    consumer.start()


if __name__ == "__main__":
    sys.exit(main())



import asyncio
import logging
import warnings
from uuid import uuid4

import msgpack
from aiokafka import AIOKafkaConsumer
from qtpy.QtCore import QObject, Signal

from haven import load_config

log = logging.getLogger(__name__)


class KafkaClient(QObject):
    run_started = Signal(str)
    run_updated = Signal(str)
    run_stopped = Signal(str)

    kafka_task = None

    def __init__(self, kafka_consumer: AIOKafkaConsumer = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._descriptors = {}
        self.kafka_consumer = kafka_consumer

    def start(self):
        # Make sure it's not already started
        if self.kafka_task is not None and not self.kafka_task.done():
            raise RuntimeError("Kafka client is already started.")
        # Start the client
        self.kafka_task = asyncio.ensure_future(self.consumer_loop())

    async def stop(self):
        await self.kafka_consumer.stop()

    async def consumer_loop(self):
        # Create a kafka consumer if one was not provided
        if self.kafka_consumer is None:
            config = load_config()
            self.kafka_consumer = AIOKafkaConsumer(
                config["queueserver"]["kafka_topic"],
                bootstrap_servers="fedorov.xray.aps.anl.gov:9092",
                group_id=str(uuid4()),
                value_deserializer=msgpack.loads,
            )
        consumer = self.kafka_consumer
        # Get cluster layout and join group `my-group`
        await consumer.start()
        # Consume events from the queueserver
        async for record in consumer:
            try:
                doc_type, doc = record.value
                log.debug(f"Received kafka record. {doc_type=}")
                self._process_document(doc_type, doc)
            except Exception as ex:
                log.exception(ex)
                continue

    def _descriptor_to_run_uid(self, descriptor_uid):
        descriptor = self._descriptors[descriptor_uid]
        run_uid = descriptor["run_start"]
        return run_uid

    def _drop_descriptors(self, run_uid: str):
        self._descriptors = {
            uid: doc
            for uid, doc in self._descriptors.items()
            if doc["run_start"] != run_uid
        }

    def _process_document(self, doc_type, doc):
        if doc_type == "start":
            # Notify clients that a new run has started
            uid = doc.get("uid", "")
            log.info(f"Received new start UID: {uid}")
            print(f"Emitting run start: {uid=}")
            self.run_started.emit(uid)
        elif doc_type == "descriptor":
            # Save the description to reference to later
            self._descriptors[doc["uid"]] = doc
        elif doc_type == "event":
            # Notify clients that this run has a new event
            descriptor_uid = doc.get("descriptor", "")
            try:
                run_uid = self._descriptor_to_run_uid(descriptor_uid)
            except KeyError:
                log.warning("fUnknown descriptor UID {descriptor_uid}")
                return
            log.info(f"Emitting run updated: {run_uid=}")
            self.run_updated.emit(run_uid)
        elif doc_type == "stop":
            run_uid = doc["run_start"]
            self._drop_descriptors(run_uid)
            print(f"Emitting run stopped: {run_uid=}")
            self.run_stopped.emit(run_uid)
        else:
            warnings.warn(f"Unknown document type '{doc_type}'")



from collections import namedtuple
from pathlib import Path

import numpy as np
from qtpy import QtCore, QtWidgets, uic
from qtpy.QtCore import Signal

axes = namedtuple("axes", ("z", "y", "x"))


class XRFView(QtWidgets.QWidget):
    ui_file = Path(__file__).parent / "xrf_view.ui"
    spectra: np.ndarray | None = None
    aggregators = {
        "Mean": np.mean,
        "Median": np.median,
        "StDev": np.std,
    }

    dataset_changed = QtCore.Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.ui = uic.loadUi(self.ui_file, self)
        self.ui.dataset_combobox.currentTextChanged.connect(self.dataset_changed)
        # Clear rows from the dimension layout
        for row in range(1, self.row_count(self.ui.dimensions_layout)):
            self.remove_dimension_widgets(row)
        # Set up some button groups for the radio buttons
        self.x_group = QtWidgets.QButtonGroup()
        self.y_group = QtWidgets.QButtonGroup()
        self.z_group = QtWidgets.QButtonGroup()
        # Other UI updates
        self.z_combobox.addItems(["None"] + list(self.aggregators.keys()))

    def row_count(self, layout: QtWidgets.QGridLayout) -> int:
        """How many rows in *layout* actually contain widgets."""
        rows = 0
        for i in range(layout.count()):
            row, _, span, _ = layout.getItemPosition(i)
            rows = max(rows, row + span)
        return rows

    def remove_dimension_widgets(self, row_idx: int):
        layout = self.ui.dimensions_layout
        for col in range(layout.columnCount()):
            widget = layout.itemAtPosition(row_idx, col)
            print(row_idx, col, widget.widget())
            layout.removeItem(widget)

    def add_dimension_widgets(self, row_idx: int, shape: int):
        layout = self.ui.dimensions_layout

        dim_label = QtWidgets.QLabel()
        dim_label.setText(str(row_idx - 1))
        layout.addWidget(dim_label, row_idx, 0)
        shape_label = QtWidgets.QLabel()
        shape_label.setText(str(shape))
        layout.addWidget(shape_label, row_idx, 1)
        # Dimension radio buttons
        z_radio = QtWidgets.QRadioButton()
        self.z_group.addButton(z_radio)
        layout.addWidget(z_radio, row_idx, 2)
        y_radio = QtWidgets.QRadioButton()
        self.y_group.addButton(y_radio)
        layout.addWidget(y_radio, row_idx, 3)
        x_radio = QtWidgets.QRadioButton()
        self.x_group.addButton(x_radio)
        layout.addWidget(x_radio, row_idx, 4)
        # Combobox for aggregation
        combobox = QtWidgets.QComboBox()
        layout.addWidget(combobox, row_idx, 5)
        combobox.addItems(self.aggregators.keys())

    def update_signal_widgets(
        self, data_keys: dict, independent_hints: [], dependent_hints: []
    ):
        """Update the UI based on new data keys and hints."""
        self.independent_hints = independent_hints
        self.dependent_hints = dependent_hints
        self.data_keys = {
            key: props for key, props in data_keys.items() if props["dtype"] == "array"
        }
        combobox = self.ui.dataset_combobox
        combobox.clear()
        combobox.addItems(self.data_keys.keys())

    def plot_spectra(self, spectra: np.ndarray | None = None):
        """Plot a set of spectra as lines.

        If *spectra* is not given, the last spectra seen are used.

        """
        if spectra is None:
            spectra = self.spectra
        else:
            self.spectra = spectra
        spectra = self.reduce_dimensions(spectra)
        # Plot the spectra
        plot = self.ui.xrf_spectra_view.getPlotItem()
        plot.clear()
        for spectrum in spectra:
            plot.plot(spectrum)

    def update_dimension_widgets(self, spectra: np.ndarray | None = None):
        """Update the widgets for setting dimensions to match the
        *spectra* array.

        """
        row_count = self.row_count(self.ui.dimensions_layout) - 1
        dim_count = spectra.ndim
        for row in range(dim_count, row_count):
            self.remove_dimension_widgets(row + 1)
        for row in range(row_count, dim_count):
            self.add_dimension_widgets(row + 1, shape=spectra.shape[row])

    def reduce_dimensions(self, spectra: np.ndarray) -> np.ndarray:
        """Reduce the input *spectra* to a shape for plotting
        depending on the dimensions widgets.

        """
        layout = self.ui.dimensions_layout
        ndim = spectra.ndim
        # Go in reverse to we can apply aggregators without messing up earlier indices
        for dim in reversed(range(spectra.ndim)):
            row = dim + 1
            is_plotted = any(
                [
                    layout.itemAtPosition(row, col).widget().isChecked()
                    for col in range(2, 5)
                ]
            )
            if not is_plotted:
                agg_name = layout.itemAtPosition(row, 5).widget().currentText()
                aggregator = self.aggregators[agg_name]
                spectra = aggregator(spectra, axis=dim)
        # Order axes according to (z, x, y) widgets
        frm, to = [], []
        src_dim = 0
        for dim in range(ndim):
            row = dim + 1
            btns = [layout.itemAtPosition(row, col).widget() for col in range(2, 5)]
            is_checked = [btn.isChecked() for btn in btns]
            try:
                plot_idx = is_checked.index(True)
            except ValueError:
                continue
            to.append(plot_idx)
            frm.append(src_dim)
            src_dim += 1
        spectra = np.moveaxis(spectra, frm, to)
        # Reduce the z-axis if plotting in 2D
        if not self.ui.z_checkbox.checkState():
            aggregator = self.aggregators[self.ui.z_combobox.currentText()]
            spectra = aggregator(spectra, axis=0)
        return spectra



from pathlib import Path

from ophyd_async.core import UUIDFilenameProvider, YMDPathProvider

from ..._iconfig import load_config


class HavenDetector: ...


def default_path_provider(path: Path = None, config=None):
    if config is None:
        config = load_config()
    if path is None:
        path = Path(config.get("area_detector_root_path", "/tmp"))
    path_provider = YMDPathProvider(
        filename_provider=UUIDFilenameProvider(),
        base_directory_path=path,
        create_dir_depth=-4,
    )
    return path_provider



from unittest.mock import MagicMock

import pandas as pd
import pytest

from firefly.run_browser.widgets import Browser1DPlotWidget


@pytest.fixture()
def plot_1d_widget(qtbot):
    widget = Browser1DPlotWidget()
    qtbot.addWidget(widget)
    return widget


def test_plot_1d_runs(plot_1d_widget):
    widget = plot_1d_widget
    assert len(widget.data_items) == 0
    # Set some runs
    widget.plot_runs({"hello": pd.Series(data=[10, 20, 30], index=[1, 2, 3])})
    assert "hello" in widget.data_items.keys()
    # Now update it again and check that the data item is reused
    mock_data_item = MagicMock()
    widget.data_items["hello"] = mock_data_item
    widget.plot_runs({"hello": pd.Series(data=[40, 50], index=[4, 5])})
    assert widget.data_items["hello"] is mock_data_item
    mock_data_item.setData.assert_called_once()


def test_clear_1d_plot(plot_1d_widget):
    widget = plot_1d_widget
    assert len(widget.data_items) == 0
    # Set some runs
    widget.plot_runs({"hello": pd.Series(data=[10, 20, 30], index=[1, 2, 3])})
    # Check that the right things get drawn
    plot_item = widget.getPlotItem()
    line = widget.cursor_line
    assert line is not None
    assert line in plot_item.items
    # Clear the plot
    widget.clear_runs()
    assert widget.cursor_line is None
    assert line not in plot_item.items
    assert len(plot_item.items) == 0


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from haven import sanitize_name


def test_sanitize_name():
    assert (
        sanitize_name("high_heat_load_mirror_1-transverse")
        == "high_heat_load_mirror_1_transverse"
    )
    assert (
        sanitize_name("high_heat_load_mirror_1.transverse")
        == "high_heat_load_mirror_1_transverse"
    )


def test_sanitize_name_with_spaces():
    # Some devices use the .DESC field, might have spaces, too
    assert sanitize_name("Det InOut") == "Det_InOut"



"""Ophyd device support for a set of XIA PFCU-controlled filters.

A PFCUFilterBank controls a set of 4 filters. Optionally, 2 filters in
a filter bank can be used as a shutter.

"""

import logging
from enum import IntEnum
from typing import Sequence

from ophyd_async.core import (
    DeviceVector,
    StandardReadable,
    StrictEnum,
    SubsetEnum,
    soft_signal_rw,
)
from ophyd_async.epics.core import epics_signal_r, epics_signal_rw

from haven.devices.shutter import ShutterState
from haven.devices.signal import derived_signal_r, derived_signal_rw
from haven.positioner import Positioner

__all__ = ["PFCUFilterBank", "PFCUFilter", "PFCUShutter"]

log = logging.getLogger(__name__)


class ConfigBits(StrictEnum):
    ZERO = "0000"
    ONE = "0001"
    TWO = "0010"
    THREE = "0011"
    FOUR = "0100"
    FIVE = "0101"
    SIX = "0110"
    SEVEN = "0111"
    EIGHT = "1000"
    NINE = "1001"
    TEN = "1010"
    ELEVEN = "1011"
    TWELVE = "1100"
    THIRTEEN = "1101"
    FOURTEEN = "1110"
    FIFTEEN = "1111"


class FilterPosition(SubsetEnum):
    OUT = "Out"
    IN = "In"
    SHORT_CIRCUIT = "Short Circuit"
    OPEN_CIRCUIT = "Open Circuit"


class FilterState(IntEnum):
    OUT = 0  # 0b000
    IN = 1  # 0b001
    FAULT = 3  # 0b011
    UNKNOWN = 4  # 0b100


class Material(SubsetEnum):
    ALUMINUM = "Al"
    MOLYBDENUM = "Mo"
    TITANIUM = "Ti"
    GLASS = "Glass"
    OTHER = "Other"


def normalize_readback(values, readback):
    return {
        FilterPosition.OUT: FilterState.OUT,
        FilterPosition.IN: FilterState.IN,
        FilterPosition.SHORT_CIRCUIT: FilterState.FAULT,
        FilterPosition.OPEN_CIRCUIT: FilterState.FAULT,
    }.get(values[readback], FilterState.UNKNOWN)


class PFCUFilter(Positioner):
    """A single filter in a PFCU filter bank.

    E.g. 25idc:pfcu0:filter1_mat
    """

    _ophyd_labels_ = {"filters"}

    def __init__(self, prefix: str, *, name: str = ""):
        with self.add_children_as_readables("config"):
            self.material = epics_signal_rw(Material, f"{prefix}_mat")
            self.thick = epics_signal_rw(float, f"{prefix}_thick")
            self.thick_unit = epics_signal_rw(str, f"{prefix}_thick.EGU")
            self.notes = epics_signal_rw(str, f"{prefix}_other")
        # We need a second "private" readback to standardize the types
        self._readback = epics_signal_r(FilterPosition, f"{prefix}_RBV")
        with self.add_children_as_readables():
            self.readback = derived_signal_r(
                int,
                derived_from={"readback": self._readback},
                inverse=normalize_readback,
            )
        self.setpoint = epics_signal_rw(bool, prefix)
        # Just use convenient values for positioner signals since there's no real position
        self.velocity = soft_signal_rw(float, initial_value=0.5)
        self.units = soft_signal_rw(str, initial_value="")
        self.precision = soft_signal_rw(int, initial_value=0)
        super().__init__(name=name)


shutter_state_map = {
    # (top filter, bottom filter): state
    (ShutterState.OPEN, ShutterState.CLOSED): ShutterState.OPEN,
    (ShutterState.CLOSED, ShutterState.OPEN): ShutterState.CLOSED,
    (ShutterState.OPEN, ShutterState.OPEN): ShutterState.FAULT,
    (ShutterState.CLOSED, ShutterState.CLOSED): ShutterState.FAULT,
}


class PFCUFilterBank(StandardReadable):
    """A XIA PFCU4 bank of four filters and/or shutters.

    Filters are indexed from 0, even though the EPICS support indexes
    from 1.

    Parameters
    ==========
    shutters
      Sets of filter numbers to use as shutters. Each entry in
      *shutters* should be a tuple like (top, bottom) where the first
      filter (top) is open when the filter is set to "out".

    """

    num_slots: int

    def __init__(
        self,
        prefix: str,
        *,
        name: str = "",
        num_slots: int = 4,
        shutters: Sequence[tuple[int, int]] = [],
    ):
        all_shutters = [v for shutter in shutters for v in shutter]
        is_in_bounds = [0 < shtr < num_slots for shtr in all_shutters]
        if not all(is_in_bounds):
            raise ValueError(
                f"Shutter indices {shutters} for filterbank {name=} must be in the range (0, {num_slots})."
            )
        self.num_slots = num_slots
        # Positioner signals
        self.setpoint = epics_signal_rw(ConfigBits, f"{prefix}config")
        with self.add_children_as_readables():
            self.readback = epics_signal_r(ConfigBits, f"{prefix}config_RBV")
        # Sort out filters vs shutters
        filters = [idx for idx in range(num_slots) if idx not in all_shutters]
        with self.add_children_as_readables():
            # Create shutters
            self.shutters = DeviceVector(
                {
                    idx: PFCUShutter(
                        prefix=prefix,
                        top_filter=top,
                        bottom_filter=btm,
                        filter_bank=self,
                    )
                    for idx, (top, btm) in enumerate(shutters)
                }
            )
            # Create filters
            self.filters = DeviceVector(
                {idx: PFCUFilter(prefix=f"{prefix}filter{idx+1}") for idx in filters}
            )
        super().__init__(name=name)


class PFCUShutter(Positioner):
    """A shutter made of two PFCU4 filters.

    For faster operation, both filters will be moved at the same
    time. That means this shutter must be a sub-component of a
    :py:class:`PFCUFilterBank`.

    Parameters
    ==========
    top_filter
      The index of the filter that is open when the filter is set to
      "out".
    bottom_filter
      The index of the filter that is open when the filter is set to
      "in".
    filter_bank
      The parent filter bank that this shutter is a part of. This
      *filter_bank*'s *setpoint* and *readback* signals will be used
      for actuating both shutter blades together.

    """

    _ophyd_labels_ = {"shutters", "fast_shutters"}

    def __init__(
        self,
        prefix: str,
        *,
        name: str = "",
        top_filter: int,
        bottom_filter: int,
        filter_bank: PFCUFilterBank,
        **kwargs,
    ):
        self._top_filter_idx = top_filter
        self._bottom_filter_idx = bottom_filter
        with self.add_children_as_readables():
            self.bottom_filter = PFCUFilter(prefix=f"{prefix}filter{bottom_filter+1}")
            self.top_filter = PFCUFilter(prefix=f"{prefix}filter{top_filter+1}")
        # Set up the positioner signals
        parent_signals = {
            "setpoint": filter_bank.setpoint,
            "readback": filter_bank.readback,
        }
        self.setpoint = derived_signal_rw(
            int, derived_from=parent_signals, forward=self.forward, inverse=self.inverse
        )
        with self.add_children_as_readables():
            self.readback = derived_signal_r(
                int, derived_from=parent_signals, inverse=self.inverse
            )
        # Just use convenient values for positioner signals since there's no real position
        self.velocity = soft_signal_rw(float, initial_value=0.5)
        self.units = soft_signal_rw(str, initial_value="")
        self.precision = soft_signal_rw(int, initial_value=0)
        super().__init__(
            name=name,
            put_complete=True,
            **kwargs,
        )

    async def forward(self, value, setpoint, readback):
        """Convert shutter state to filter bank state."""
        # Bit masking to set both blades together
        readback_value = await readback.get_value()
        num_bits = len(readback_value)
        old_bits = int(readback_value, 2)
        if value == ShutterState.OPEN:
            open_bits = self.bottom_mask()
            close_bits = self.top_mask()
        elif value == ShutterState.CLOSED:
            close_bits = self.bottom_mask()
            open_bits = self.top_mask()
        else:
            raise ValueError(bin(value))
        new_bits = (old_bits | open_bits) & (0b1111 - close_bits)
        log.debug(f"{old_bits=:b}, {open_bits=:b}, {close_bits=:b}, {new_bits=:b}")
        return {setpoint: f"{new_bits:0b}".zfill(num_bits)}

    def inverse(self, values, readback, **kwargs):
        """Convert filter bank state to shutter state."""
        bits = int(values[readback], 2)
        # Determine which filters are open and closed
        top_position = int(bool(bits & self.top_mask()))
        bottom_position = int(bool(bits & self.bottom_mask()))
        result = shutter_state_map[(top_position, bottom_position)]
        return result

    def _mask(self, pos):
        num_filters = 4
        return 1 << (num_filters - pos - 1)

    def top_mask(self):
        return self._mask(self._top_filter_idx)

    def bottom_mask(self):
        return self._mask(self._bottom_filter_idx)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
import warnings
from enum import IntEnum, unique
from typing import Mapping

from ophyd.utils.errors import ReadOnlyError
from ophyd_async.core import soft_signal_rw
from ophyd_async.epics.core import epics_signal_r

from ..positioner import Positioner
from .signal import derived_signal_rw, epics_signal_xval

__all__ = ["PssShutter", "ShutterState"]


log = logging.getLogger(__name__)


@unique
class ShutterState(IntEnum):
    OPEN = 0  # 0b000
    CLOSED = 1  # 0b001
    FAULT = 3  # 0b011
    UNKNOWN = 4  # 0b100


class PssShutter(Positioner):
    _ophyd_labels_ = {"shutters"}
    _last_setpoint: int = ShutterState.UNKNOWN
    allow_open: bool
    allow_close: bool

    def __init__(
        self,
        prefix: str,
        name: str,
        allow_open: bool = True,
        allow_close: bool = True,
        labels={"shutters"},
        **kwargs,
    ):
        self.allow_open = allow_open
        self.allow_close = allow_close
        # Actuators for opening/closing the shutter
        self.open_signal = epics_signal_xval(f"{prefix}OpenEPICSC")
        self.close_signal = epics_signal_xval(f"{prefix}CloseEPICSC")
        # Just use convenient values for these since there's no real position
        self.velocity = soft_signal_rw(float, initial_value=0.5)
        self.units = soft_signal_rw(str, initial_value="")
        self.precision = soft_signal_rw(int, initial_value=0)
        # Positioner signals for moving the shutter
        with self.add_children_as_readables():
            self.readback = epics_signal_r(bool, f"{prefix}BeamBlockingM.VAL")
        self.setpoint = derived_signal_rw(
            int,
            derived_from={
                "open_signal": self.open_signal,
                "close_signal": self.close_signal,
            },
            forward=self._actuate_shutter,
            inverse=self._shutter_setpoint,
        )
        super().__init__(name=name, **kwargs)

    def check_value(self, pos):
        """Check that the shutter has the right permissions."""
        if pos == ShutterState.CLOSED and not self.allow_close:
            raise ReadOnlyError(
                f"Shutter {self.name} is not permitted to be closed. Set `allow_close` for this shutter."
            )
        if pos == ShutterState.OPEN and not self.allow_open:
            raise ReadOnlyError(
                f"Shutter {self.name} is not permitted to be opened per iconfig.toml. Set `allow_open` for this shutter."
            )

    async def _actuate_shutter(self, value: int, open_signal, close_signal) -> Mapping:
        """Open/close the shutter using derived-from signals."""
        self.check_value(value)
        if value == ShutterState.OPEN:
            items = {open_signal: 1}
        elif value == ShutterState.CLOSED:
            items = {close_signal: 1}
        else:
            raise ValueError(f"Invalid shutter state for {self}")
        return items

    def _shutter_setpoint(self, values: Mapping, open_signal, close_signal) -> int:
        """Determine whether the shutter was last opened or closed."""
        do_open = values[open_signal]
        do_close = values[close_signal]
        if do_open and do_close:
            # Shutter is both opening and closing??
            warnings.warn("Unknown shutter setpoint")
            self._last_setpoint = ShutterState.UNKNOWN
        elif do_open:
            self._last_setpoint = ShutterState.OPEN
        elif do_close:
            self._last_setpoint = ShutterState.CLOSED
        return self._last_setpoint


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest

from haven.devices.detectors.aravis import AravisDetector

PREFIX = "255idgigeA:"


@pytest.fixture()
async def camera(sim_registry):
    camera = AravisDetector(prefix=PREFIX, name="s255id-gige-A")
    await camera.connect(mock=True)
    sim_registry.register(camera, labels={"cameras", "detectors"})
    # Registry with the simulated registry
    return camera


@pytest.mark.asyncio
async def test_camera_trigger_source_choices(camera):
    """Confirm the camera device has the right signals.

    Solves this inital error received when using detaul AravisDetector.

    > drv: NotConnected:
    >     trigger_source: TypeError: 25idcARV4:cam1:TriggerSource_RBV has choices ('Software', 'Line1', 'Line3', 'Action1'), which is not a superset of SubsetEnum['Freerun', 'Line1'].
    >     data_type: NotConnected: ca://25idcARV4:cam1:NDDataType_RBV
    > hdf: NotConnected:
    >     data_type: NotConnected: ca://25idcARV4:HDF1:NDDataType_RBV

    """
    desc = await camera.drv.trigger_source.describe()
    choices = desc["s255id-gige-A-drv-trigger_source"]["choices"]
    assert "Software" in choices
    assert "Line1" in choices


@pytest.mark.asyncio
async def test_camera_signals(camera):
    """Confirm the camera device has the right signals.

    Solves this inital error received when using detaul AravisDetector.

    > drv: NotConnected:
    >     trigger_source: TypeError: 25idcARV4:cam1:TriggerSource_RBV has choices ('Software', 'Line1', 'Line3', 'Action1'), which is not a superset of SubsetEnum['Freerun', 'Line1'].
    >     data_type: NotConnected: ca://25idcARV4:cam1:NDDataType_RBV
    > hdf: NotConnected:
    >     data_type: NotConnected: ca://25idcARV4:HDF1:NDDataType_RBV

    """
    desc = await camera.drv.data_type.describe()
    cam_source = desc["s255id-gige-A-drv-data_type"]["source"]
    assert cam_source == "mock+ca://255idgigeA:cam1:DataType_RBV"
    # Check HDF signal source
    desc = await camera.hdf.data_type.describe()
    hdf_source = desc["s255id-gige-A-hdf-data_type"]["source"]
    assert hdf_source == "mock+ca://255idgigeA:HDF1:DataType_RBV"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from .baseline import baseline_decorator, baseline_wrapper  # noqa: F401
from .inject_metadata import inject_haven_md_wrapper  # noqa: F401
from .open_shutters import open_shutters_decorator, open_shutters_wrapper  # noqa: F401
from .shutter_suspender import (  # noqa: F401
    shutter_suspend_decorator,
    shutter_suspend_wrapper,
)



import asyncio
import xml.etree.ElementTree as ET
from collections.abc import Sequence

import numpy as np
from ophyd_async.core import (
    Array1D,
    AsyncStatus,
    DetectorController,
    Device,
    DeviceVector,
    PathProvider,
    SignalR,
    StandardDetector,
    StrictEnum,
    TriggerInfo,
    soft_signal_r_and_setter,
)
from ophyd_async.epics import adcore
from ophyd_async.epics.adcore._utils import (
    ADBaseDataType,
    NDAttributeDataType,
    NDAttributeParam,
    convert_ad_dtype_to_np,
)
from ophyd_async.epics.core import epics_signal_r, epics_signal_rw, epics_signal_x

from .area_detectors import HavenDetector, default_path_provider


class XspressTriggerMode(StrictEnum):
    SOFTWARE = "Software"
    INTERNAL = "Internal"
    IDC = "IDC"
    TTL_VETO_ONLY = "TTL Veto Only"
    TTL_BOTH = "TTL Both"
    LVDS_VETO_ONLY = "LVDS Veto Only"
    LVDS_BOTH = "LVDS Both"
    SOFTWARE_INTERNAL = "Software + Internal"


class XspressDriverIO(adcore.ADBaseIO):
    def __init__(self, prefix, name=""):
        self.trigger_mode = epics_signal_rw(XspressTriggerMode, f"{prefix}TriggerMode")
        self.erase_on_start = epics_signal_rw(bool, f"{prefix}EraseOnStart")
        self.erase = epics_signal_x(f"{prefix}ERASE")
        self.deadtime_correction = epics_signal_rw(bool, f"{prefix}CTRL_DTC")
        self.number_of_elements = epics_signal_r(int, f"{prefix}MaxSizeY_RBV")
        super().__init__(prefix=prefix, name=name)


class XspressController(DetectorController):
    def __init__(self, driver: adcore.ADBaseIO) -> None:
        self._drv = driver

    def get_deadtime(self, exposure: float) -> float:
        # Arbitrary value. To-do: fill this in when we know what to
        # include
        return 0.001

    async def setup_ndattributes(self, device_name: str):
        num_elements = await self._drv.number_of_elements.get_value()
        params = ndattribute_params(
            device_name=device_name, elements=range(num_elements)
        )
        xml = ndattribute_xml(params)
        await self._drv.nd_attributes_file.set(xml)

    @AsyncStatus.wrap
    async def prepare(self, trigger_info: TriggerInfo):
        await asyncio.gather(
            self._drv.num_images.set(trigger_info.total_number_of_triggers),
            self._drv.image_mode.set(adcore.ImageMode.MULTIPLE),
            self._drv.trigger_mode.set(XspressTriggerMode.INTERNAL),
            # Hardware deadtime correciton is not reliable
            # https://github.com/epics-modules/xspress3/issues/57
            self._drv.deadtime_correction.set(False),
        )

    async def wait_for_idle(self):
        if self._arm_status:
            await self._arm_status

    async def arm(self):
        self._arm_status = await adcore.start_acquiring_driver_and_ensure_status(
            self._drv
        )

    async def disarm(self):
        await adcore.stop_busy_record(self._drv.acquire, False, timeout=1)


class XspressDatasetDescriber(adcore.ADBaseDatasetDescriber):
    """The datatype cannot be reliably determined from DataType_RBV.

    Instead, read out whether deadtime correction is enabled and
    determine the datatype this way.

    https://github.com/epics-modules/xspress3/issues/57

    """

    async def np_datatype(self) -> str:
        dt_correction = await self._driver.deadtime_correction.get_value()
        if dt_correction:
            return convert_ad_dtype_to_np(ADBaseDataType.FLOAT64)
        else:
            return convert_ad_dtype_to_np(ADBaseDataType.UINT32)


class XspressElement(Device):
    """Data and controls for an individual Xspress3 detector element."""

    def __init__(self, prefix: str, element_index: int, name: str = ""):
        """Parameters
        ==========
        prefix
          The detector's overall PV prefix, not including the
          e.g. "MCA1:" part.
        element_index
          This elements positional index, e.g. the first element will
          have *element_index*=0.

        """
        elem_num = element_index + 1
        self.spectrum = epics_signal_r(
            Array1D[np.float64], f"{prefix}MCA{elem_num}:ArrayData"
        )
        self.dead_time_percent = epics_signal_r(
            float, f"{prefix}C{elem_num}SCA:10:Value_RBV"
        )
        self.dead_time_factor = epics_signal_r(
            float, f"{prefix}C{elem_num}SCA:9:Value_RBV"
        )
        super().__init__(name=name)


class Xspress3Detector(HavenDetector, StandardDetector):
    """A detector controlled by Xspress3 electronics.

    The elements of the detector are represented on the *mcas*
    attribute. The number of mcas is determined by passing *elements*
    when initializing an object. *elements* can be either an integer,
    in which case it represents to number of elements, or an iterator
    representing the index of each element.

    The following lines are equivalent.

    .. code-block:: python

      det = Xspress3Detector(..., elements=4)

      det = Xspress3Detector(..., elements=[0, 1, 2, 3])

      det = Xspress3Detector(..., elements=range(4))

    The parameter *ev_per_bin* controls the conversion from histogram
    bin to energy. This is set during calibration; changing the value
    passed into this class without re-calibrating the detector **will
    result in an incorrect conversion**.

    """

    _ophyd_labels_ = {"detectors", "xrf_detectors"}
    _controller: DetectorController
    _writer: adcore.ADHDFWriter

    def __init__(
        self,
        prefix: str,
        path_provider: PathProvider | None = None,
        elements: int | Sequence[int] = 1,
        ev_per_bin: float = 10.0,
        drv_suffix="det1:",
        hdf_suffix="HDF1:",
        name: str = "",
        config_sigs: Sequence[SignalR] = (),
    ):
        # Per-element MCA devices
        try:
            elements = range(elements)
        except TypeError:
            pass
        self.elements = DeviceVector(
            {idx: XspressElement(prefix, element_index=idx) for idx in elements}
        )
        # Area detector IO devices
        self.drv = XspressDriverIO(prefix + drv_suffix)
        self.hdf = adcore.NDFileHDFIO(prefix + hdf_suffix)

        if path_provider is None:
            path_provider = default_path_provider()
        # Extra configuration signals
        self.ev_per_bin, _ = soft_signal_r_and_setter(float, initial_value=ev_per_bin)

        super().__init__(
            XspressController(self.drv),
            adcore.ADHDFWriter(
                self.hdf,
                path_provider,
                lambda: self.name,
                XspressDatasetDescriber(self.drv),
                self.drv,  # <- for DT ndattributes
            ),
            config_sigs=(
                self.drv.acquire_period,
                self.drv.acquire_time,
                self.ev_per_bin,
                *config_sigs,
            ),
            name=name,
        )

    @AsyncStatus.wrap
    async def stage(self) -> None:
        self._old_xml_file, *_ = await asyncio.gather(
            self.drv.nd_attributes_file.get_value(),
            super().stage(),
            self._controller.setup_ndattributes(device_name=self.name),
            self.drv.erase_on_start.set(False),
            self.drv.erase.trigger(),
        )

    @AsyncStatus.wrap
    async def unstage(self) -> None:
        await super().unstage()
        if self._old_xml_file is not None:
            # Restore the original XML attributes file
            await self.drv.nd_attributes_file.set(self._old_xml_file)
            self._old_xml_file = None

    @property
    def default_time_signal(self):
        return self.drv.acquire_time


def ndattribute_xml(params):
    """Convert a set of NDAttribute params to XML."""
    root = ET.Element("Attributes")
    for ndattribute in params:
        ET.SubElement(
            root,
            "Attribute",
            name=ndattribute.name,
            type="PARAM",
            source=ndattribute.param,
            addr=str(ndattribute.addr),
            datatype=ndattribute.datatype.value,
            description=ndattribute.description,
        )
    xml_text = ET.tostring(root, encoding="unicode")
    return xml_text


def ndattribute_params(
    device_name: str, elements: Sequence[int]
) -> Sequence[NDAttributeParam]:
    """Create a set of ndattribute params that can be written to the AD's
    HDF5 file.

    These parameters can then be used with something like
    :py:func:`ophyd_async.plan_stubs.setup_ndattributes` to build the
    XML.

    """
    params = []
    for idx in elements:
        new_params = [
            NDAttributeParam(
                name=f"{device_name}-element{idx}-deadtime_factor",
                param="XSP3_CHAN_DTFACTOR",
                datatype=NDAttributeDataType.DOUBLE,
                addr=idx,
                description=f"Chan {idx} DTC Factor",
            ),
            NDAttributeParam(
                name=f"{device_name}-element{idx}-deadtime_percent",
                param="XSP3_CHAN_DTPERCENT",
                datatype=NDAttributeDataType.DOUBLE,
                addr=idx,
                description=f"Chan {idx} DTC Percent",
            ),
            NDAttributeParam(
                name=f"{device_name}-element{idx}-event_width",
                param="XSP3_EVENT_WIDTH",
                datatype=NDAttributeDataType.DOUBLE,
                addr=idx,
                description=f"Chan {idx} Event Width",
            ),
            NDAttributeParam(
                name=f"{device_name}-element{idx}-clock_ticks",
                param="XSP3_CHAN_SCA0",
                datatype=NDAttributeDataType.DOUBLE,
                addr=idx,
                description=f"Chan {idx} ClockTicks",
            ),
            NDAttributeParam(
                name=f"{device_name}-element{idx}-reset_ticks",
                param="XSP3_CHAN_SCA1",
                datatype=NDAttributeDataType.DOUBLE,
                addr=idx,
                description=f"Chan {idx} ResetTicks",
            ),
            NDAttributeParam(
                name=f"{device_name}-element{idx}-reset_counts",
                param="XSP3_CHAN_SCA2",
                datatype=NDAttributeDataType.DOUBLE,
                addr=idx,
                description=f"Chan {idx} ResetCounts",
            ),
            NDAttributeParam(
                name=f"{device_name}-element{idx}-all_event",
                param="XSP3_CHAN_SCA3",
                datatype=NDAttributeDataType.DOUBLE,
                addr=idx,
                description=f"Chan {idx} AllEvent",
            ),
            NDAttributeParam(
                name=f"{device_name}-element{idx}-all_good",
                param="XSP3_CHAN_SCA4",
                datatype=NDAttributeDataType.DOUBLE,
                addr=idx,
                description=f"Chan {idx} AllGood",
            ),
            NDAttributeParam(
                name=f"{device_name}-element{idx}-pileup",
                param="XSP3_CHAN_SCA7",
                datatype=NDAttributeDataType.DOUBLE,
                addr=idx,
                description=f"Chan {idx} Pileup",
            ),
        ]
        params.extend(new_params)
    return params



import numpy as np
import pytest
from qtpy.QtWidgets import QComboBox

from firefly.run_browser.xrf_view import XRFView


@pytest.fixture()
def view(qtbot):
    xrf_view = XRFView()
    qtbot.addWidget(xrf_view)
    return xrf_view


def test_load_ui(view):
    """Make sure widgets were loaded from the UI file."""
    assert isinstance(view.ui.dataset_combobox, QComboBox)


def test_dataset_combobox_options(view):
    """
    We need to know:
    - external signals
    - internal signals
    - hints

    Then each view can handle sorting out which signals it needs.

    Use '64e85e20-106c-48e6-b643-77e9647b0242' for testing in the
    haven-dev catalog.

    """

    data_keys = {
        "ge_8element": {
            "dtype": "array",
            "dtype_numpy": "<u4",
            "external": "STREAM:",
            "object_name": "ge_8element",
            "shape": [8, 4096],
            "source": "ca://XSP_Ge_8elem:HDF1:FullFileName_RBV",
        },
        "I0-net_current": {
            "dtype": "number",
            "dtype_numpy": "<f8",
            "object_name": "I0",
            "shape": [],
            "source": "ca://25idVME:3820:scaler1.S0",
            "units": "A",
        },
    }
    ihints = ["sim_motor_2"]
    dhints = ["I0-net_current"]
    combobox = view.ui.dataset_combobox
    view.update_signal_widgets(data_keys, ihints, dhints)
    assert (
        combobox.findText("ge_8element") > -1
    ), f"ge_8element signal not in {combobox.objectName()}."
    assert (
        combobox.findText("I0-net_current") == -1
    ), f"I0-net_current signal should not be in {combobox.objectName()}."


def test_update_dimension_widgets(view):
    data = np.mgrid[0:21, 0:8, 0:4096][0]
    view.update_dimension_widgets(data)
    layout = view.ui.dimensions_layout
    assert view.row_count(layout) == 4
    assert layout.itemAtPosition(1, 1).widget().text() == "21"
    assert layout.itemAtPosition(2, 1).widget().text() == "8"
    assert layout.itemAtPosition(3, 1).widget().text() == "4096"


def test_reduce_dimensions_3d(view):
    data = np.mgrid[0:14, 0:21, 0:8, 0:4096][0]
    # Make sure the dimension widgets are set up
    view.update_dimension_widgets(data)
    layout = view.ui.dimensions_layout
    z_pos, y_pos, x_pos = (2, 3, 4)
    layout.itemAtPosition(3, z_pos).widget().setChecked(True)
    layout.itemAtPosition(4, y_pos).widget().setChecked(True)
    layout.itemAtPosition(1, x_pos).widget().setChecked(True)
    # Apply dimensionality reduction
    new_data = view.reduce_dimensions(data)
    # Check results
    assert new_data.ndim == 3
    assert new_data.shape == (8, 4096, 14)


def test_reduce_dimensions_2d(view):
    data = np.mgrid[0:14, 0:21, 0:8, 0:4096][0]
    # Make sure the dimension widgets are set up
    view.update_dimension_widgets(data)
    layout = view.ui.dimensions_layout
    z_pos, y_pos, x_pos = (2, 3, 4)
    layout.itemAtPosition(3, z_pos).widget().setChecked(True)
    layout.itemAtPosition(4, y_pos).widget().setChecked(True)
    layout.itemAtPosition(1, x_pos).widget().setChecked(True)
    view.ui.z_combobox.setCurrentText("Mean")
    view.ui.z_checkbox.setChecked(False)
    # Apply dimensionality reduction
    new_data = view.reduce_dimensions(data)
    # Check results
    assert new_data.ndim == 2
    assert new_data.shape == (4096, 14)



"""
Provide beamline configuration from the iconfig.toml file.

Example TOML configuration file: iconfig_default.toml

"""

__all__ = [
    "load_config",
]

import argparse
import logging
import os
from pathlib import Path
from pprint import pprint
from typing import Optional, Sequence

import tomli
from mergedeep import merge

log = logging.getLogger(__name__)


_local_overrides = {}


def load_files(file_paths: Sequence[Path]):
    """Generate the configs for files as dictionaries."""
    for fp in file_paths:
        fp = Path(fp)
        if fp.exists():
            with open(fp, mode="rb") as fp:
                log.debug(f"Loading config file: {fp}")
                config = tomli.load(fp)
                yield config
        else:
            log.debug(f"Could not find config file, skipping: {fp}")


def lookup_file_paths():
    if os.environ.get("HAVEN_CONFIG_FILES", "") != "":
        return [Path(fp) for fp in os.environ["HAVEN_CONFIG_FILES"].split(",")]
    else:
        return [Path(__file__).parent / "iconfig_testing.toml"]


def load_config(file_paths: Optional[Sequence[Path]] = None):
    """Load TOML config files.

    Will load files specified in the following locations:

    1. *file_paths* argument
    2. The $HAVEN_CONFIG_FILES environmental variable.
    3. iconfig_default.toml file included with Haven.

    """
    if file_paths is None:
        file_paths = lookup_file_paths()
    else:
        file_paths = list(file_paths).copy()
    # Add config file from environmental variable
    # Load configuration from TOML files
    config = {}
    merge(config, *load_files(file_paths), _local_overrides)
    return config


def print_config_value(args: Sequence[str] = None):
    """Print a config value from TOML files.

    Parameters
    ----------
    key
      The path to the value to retrieve from config files. Sections
      should be separated by dots, e.g. "shutter.A.open_pv"

    """
    # Set up command line arguments
    parser = argparse.ArgumentParser(
        prog="haven_config",
        description="Retrieve a value from Haven's config files.",
    )
    parser.add_argument("key", help="The dot-separated key to look up.")
    args = parser.parse_args(args=args)
    # Get the keys from the config file
    value = load_config()
    for part in args.key.split("."):
        value = value[part]
    try:
        value = value.strip()
    except AttributeError:
        # It's not a simple string, so pretty print it
        pprint(value)
    else:
        # Simple string, so just print it
        print(value)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import re

bad_separators = re.compile("[-. ]+")


def titleize(name):
    """Convert a device name into a human-readable title."""
    title = name.replace("_", " ").title()
    # Replace select phrases that are known to be incorrect
    replacements = {"Kb ": "KB "}
    for orig, new in replacements.items():
        title = title.replace(orig, new)
    return title


def sanitize_name(name: str) -> str:
    """Convert *name* into something usable as a queueserver device."""
    return bad_separators.sub("_", name)



import asyncio
import xml.etree.ElementTree as ET
from pathlib import Path

import pytest
from ophyd_async.core import TriggerInfo
from ophyd_async.testing import get_mock_put, set_mock_value

from haven.devices.detectors.xspress import Xspress3Detector, ndattribute_params

this_dir = Path(__file__).parent


@pytest.fixture()
async def detector():
    det = Xspress3Detector("255id_xsp:", name="vortex_me4", elements=4)
    await det.connect(mock=True)
    set_mock_value(det.hdf.file_path_exists, True)
    return det


async def test_signals(detector):
    assert await detector.ev_per_bin.get_value() == 10
    # Spot-check some PVs
    assert (
        detector.drv.acquire_time.source == "mock+ca://255id_xsp:det1:AcquireTime_RBV"
    )
    assert detector.drv.acquire.source == "mock+ca://255id_xsp:det1:Acquire_RBV"
    # Individual element's signals
    assert len(detector.elements) == 4
    elem0 = detector.elements[0]
    assert elem0.spectrum.source == "mock+ca://255id_xsp:MCA1:ArrayData"
    assert elem0.dead_time_percent.source == "mock+ca://255id_xsp:C1SCA:10:Value_RBV"
    assert elem0.dead_time_factor.source == "mock+ca://255id_xsp:C1SCA:9:Value_RBV"


async def test_description(detector):
    config = await detector.read_configuration()
    assert f"{detector.name}-ev_per_bin" in config


async def test_trigger(detector):
    status = detector.trigger()
    await asyncio.sleep(0.1)  # Let the event loop turn
    set_mock_value(detector.hdf.num_captured, 1)
    await status
    # Check that signals were set
    get_mock_put(detector.drv.num_images).assert_called_once_with(1, wait=True)


async def test_stage(detector):
    assert not get_mock_put(detector.drv.erase).called
    await detector.stage()
    get_mock_put(detector.drv.erase_on_start).assert_called_once_with(False, wait=True)
    assert get_mock_put(detector.drv.erase).called


async def test_descriptor(detector):
    """There is a bug in the xspress3 EPICS driver that means it does not
    report the datatype correctly. This tests a workaround to decide
    based on the value of the dead_time_correction.

    https://github.com/epics-modules/xspress3/issues/57

    """
    # With deadtime correction off, we should get unsigned longs
    await detector.drv.deadtime_correction.set(False)
    assert await detector.writer._dataset_describer.np_datatype() == "<u4"
    # With deadtime correction on, we should get double-precision floats
    await detector.drv.deadtime_correction.set(True)
    assert await detector.writer._dataset_describer.np_datatype() == "<f8"


async def test_deadtime_correction_disabled(detector):
    """Deadtime correction in hardware is not reliable and should be
    disabled.

    https://github.com/epics-modules/xspress3/issues/57

    """
    set_mock_value(detector.drv.deadtime_correction, True)
    trigger_info = TriggerInfo(number_of_triggers=1)
    await detector.prepare(trigger_info)
    assert not await detector.drv.deadtime_correction.get_value()


def test_default_time_signal_xspress(xspress):
    # assert xspress.default_time_signal is xspress.acquire_time
    assert xspress.default_time_signal is xspress.drv.acquire_time


async def test_ndattribute_params():
    n_elem = 8
    n_params = 9
    params = ndattribute_params(device_name="xsp3", elements=range(n_elem))
    assert len(params) == n_elem * n_params


async def test_stage_ndattributes(detector):
    num_elem = 8
    set_mock_value(detector.drv.number_of_elements, num_elem)
    set_mock_value(detector.drv.nd_attributes_file, "XSP3.xml")
    await detector.stage()
    xml_mock = get_mock_put(detector.drv.nd_attributes_file)
    assert xml_mock.called
    # Test that the XML is correct
    args, kwargs = xml_mock.call_args
    tree = ET.fromstring(args[0])
    assert len(tree) == num_elem * 9
    # Check that the XML file gets reset when unstaged
    await detector.unstage()
    assert xml_mock.call_args[0][0] == "XSP3.xml"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from pathlib import Path
from unittest.mock import AsyncMock, MagicMock

import pytest

from haven.devices.ion_chamber import IonChamber
from haven.devices.motor import load_motors
from haven.devices.slits import BladeSlits
from haven.instrument import Instrument

haven_dir = Path(__file__).parent.parent.resolve()
toml_file = haven_dir / "iconfig_testing.toml"


@pytest.fixture()
def instrument():
    inst = Instrument(
        {
            "ion_chamber": IonChamber,
            "motors": load_motors,
            "blade_slits": BladeSlits,
        }
    )
    with open(toml_file, mode="tr", encoding="utf-8") as fd:
        inst.parse_toml_file(fd)
    return inst


def test_load(monkeypatch):
    instrument = Instrument({})
    # Mock out the relevant methods to test
    monkeypatch.setattr(instrument, "parse_toml_file", MagicMock())
    monkeypatch.setattr(instrument, "connect", AsyncMock(return_value=([], [])))
    monkeypatch.setenv("HAVEN_CONFIG_FILES", str(toml_file), prepend=False)
    # Execute the loading step
    instrument.load(toml_file)
    # Check that the right methods were called
    instrument.parse_toml_file.assert_called_once()



import asyncio
import functools
import logging
import os
import sqlite3
import threading
import warnings
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from typing import Sequence

import numpy as np
from tiled.client import from_uri
from tiled.client.cache import Cache

from ._iconfig import load_config

log = logging.getLogger(__name__)


def run_in_executor(_func):
    """Decorator that makes the wrapped synchronous function asynchronous.

    This is done by running the wrapped function in the default
    asyncio executor.

    """

    @functools.wraps(_func)
    def wrapped(*args, **kwargs):
        loop = asyncio.get_running_loop()
        func = functools.partial(_func, *args, **kwargs)
        return loop.run_in_executor(None, func)

    return wrapped


def unsnake(arr: np.ndarray, snaking: list) -> np.ndarray:
    """Unsnake a nump array.

    For each axis in *arr*, there should be a corresponding True/False
    in *snaking* whether that axis should have alternating rows. The
    first entry is ignored as it doesn't make sense to snake the first
    axis.

    Returns
    =======
    unsnaked
      A copy of *arr* with the odd-numbered axes flipped (if indicated
      by *snaking*).

    """
    # arr = np.copy(arr)
    # Create some slice object for easier manipulation
    full_axis = slice(None)
    alternating = slice(None, None, 2)
    flipped = slice(None, None, -1)
    # Flip each axis if necessary (skipping the first axis)
    for axis, is_snaked in enumerate(snaking[1:]):
        if not is_snaked:
            continue
        slices = (full_axis,) * axis
        slices += (alternating,)
        arr[slices] = arr[slices + (flipped,)]
    return arr


def with_thread_lock(fn):
    """Makes sure the function isn't accessed concurrently."""

    def wrapper(obj, *args, **kwargs):
        obj._lock.acquire()
        try:
            fn(obj, *args, **kwargs)
        finally:
            obj._lock.release()

    return wrapper


class ThreadSafeCache(Cache):
    """Equivalent to the regular cache, but thread-safe.

    Ensures that sqlite3 is built with concurrency features, and
    ensures that no two write operations happen concurrently.

    """

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._lock = threading.Lock()

    def write_safe(self):
        """
        Check that it is safe to write.

        SQLite is not threadsafe for concurrent _writes_.
        """
        is_main_thread = threading.current_thread().ident == self._owner_thread
        sqlite_is_safe = sqlite3.threadsafety == 1
        return is_main_thread or sqlite_is_safe

    # Wrap the accessor methods so they wait for the lock
    clear = with_thread_lock(Cache.clear)
    set = with_thread_lock(Cache.set)
    get = with_thread_lock(Cache.get)
    delete = with_thread_lock(Cache.delete)


DEFAULT_NODE = object()


def tiled_client(
    catalog: str = DEFAULT_NODE,
    uri: str = None,
    cache_filepath=None,
    structure_clients="numpy",
):
    """Load a tiled client for retrieving data from databses.

    Parameters
    ==========
    catalog
      The node within the catalog to return, by default this will be
      read from the config file. If ``None``, the root container will
      be return containing all catalogs.
    uri
      The location of the tiled server, e.g. "http://localhost:8000".
    cache_filepath
      Where to keep a local cache of tiled nodes.
    structure_clients
      "numpy" for immediate retrieval of data, "dask" for just-in-time
      retrieval.

    """
    tiled_config = load_config().get("tiled", {})
    # Create a cache for saving local copies
    if cache_filepath is None:
        cache_filepath = tiled_config.get("cache_filepath", "")
    if os.access(cache_filepath, os.W_OK):
        cache = ThreadSafeCache(filepath=cache_filepath)
    else:
        warnings.warn(f"Cache file is not writable: {cache_filepath}")
        cache = None
    # Create the client
    if uri is None:
        uri = tiled_config["uri"]
    api_key = tiled_config.get("api_key")
    client_ = from_uri(uri, structure_clients, api_key=api_key)
    if catalog is DEFAULT_NODE:
        client_ = client_[tiled_config["default_catalog"]]
    elif catalog is not None:
        client_ = client_[catalog]
    return client_


class CatalogScan:
    """A single scan from the tiled API with some convenience methods.

    Parameters
    ==========
      A tiled container on which to operate.
    """

    def __init__(self, container, executor=None):
        self.container = container
        self.executor = executor

    @run_in_executor
    def stream_names(self):
        return list(self.container.keys())

    @run_in_executor
    def _read_data(self, signals: Sequence | None, dataset: str):
        data = self.container[dataset]
        if signals is None:
            return data.read()
        # Remove duplicates and missing signals
        signals = set(signals)
        available_signals = set(data.columns)
        signals = signals & available_signals
        return data.read()

    @property
    def uid(self):
        return self.container._item["id"]

    async def export(self, filename: str, format: str):
        target = partial(self.container.export, filename, format=format)
        await self.loop.run_in_executor(None, target)

    def formats(self):
        return self.container.formats

    async def data(self, *, signals=None, stream: str = "primary"):
        return await self._read_data(signals, f"{stream}/internal/events/")

    @property
    def loop(self):
        return asyncio.get_running_loop()

    @run_in_executor
    def data_keys(self, stream: str = "primary"):
        return self.container[f"{stream}/internal/events"].columns

    async def hints(self, stream: str = "primary"):
        """Retrieve the data hints for this scan.

        Parameters
        ==========
        stream
          The name of the Tiled data stream to look up hints for.

        Returns
        =======
        independent
          The hints for the independent scanning axis.
        dependent
          The hints for the dependent scanning axis.

        """
        metadata = await self.metadata
        # Get hints for the independent (X)
        try:
            independent = metadata["start"]["hints"]["dimensions"][0][0]
        except (KeyError, IndexError):
            warnings.warn("Could not get independent hints")
        # Get hints for the dependent (X)
        dependent = []
        primary_metadata = await self._read_metadata(stream)
        hints = primary_metadata["hints"]
        for device, dev_hints in hints.items():
            dependent.extend(dev_hints["fields"])
        return independent, dependent

    @run_in_executor
    def _read_metadata(self, keys=None):
        assert keys != "", "Metadata keys cannot be ''."
        container = self.container
        if keys is not None:
            print(f"{container=}, {keys=}")
            container = container[keys]
        return container.metadata

    @property
    async def metadata(self):
        return await self._read_metadata()

    async def __getitem__(self, signal, stream: str = "primary"):
        """Retrieve a signal from the dataset, with reshaping etc."""
        arr = await self._read_data(
            [f"{stream}/{signal}"], dataset=f"{stream}/internal/events"
        )
        arr = np.asarray(arr[signal])
        # Re-shape to match the scan dimensions
        metadata = await self.metadata
        try:
            shape = metadata["start"]["shape"]
        except KeyError:
            log.warning(f"No shape found for {repr(signal)}.")
        else:
            arr = np.reshape(arr, shape)
        # Flip alternating rows if snaking is enabled
        if "snaking" in metadata["start"]:
            arr = unsnake(arr, metadata["start"]["snaking"])
        return arr


class Catalog:
    """An asynchronous wrapper around the tiled client.

    This class has a more intelligent understanding of how *our* data
    are structured, so can make some assumptions and takes care of
    boiler-plate code (e.g. reshaping maps, etc).

    Parameters
    ==========
    client
      A Tiled client that has scan UIDs as its keys.

    """

    _client = None

    def __init__(self, client=None):
        self._client = client
        self.executor = ThreadPoolExecutor()

    def __del__(self):
        self.executor.shutdown(wait=True, cancel_futures=True)

    @property
    def loop(self):
        return asyncio.get_running_loop()

    @property
    async def client(self):
        if self._client is None:
            self._client = await run_in_executor(tiled_client)()
        return self._client

    async def __getitem__(self, uid) -> CatalogScan:
        client = await self.client
        container = await run_in_executor(client.__getitem__)(uid)
        scan = CatalogScan(container=container, executor=self.executor)
        return scan

    async def items(self):
        client = await self.client
        for key, value in await run_in_executor(client.items)():
            yield key, CatalogScan(container=value, executor=self.executor)

    async def values(self):
        client = await self.client
        containers = await run_in_executor(client.values)()
        for container in containers:
            yield CatalogScan(container, executor=self.executor)

    async def __len__(self):
        client = await self.client
        length = await run_in_executor(client.__len__)()
        return length

    async def search(self, query):
        """
        Make a Node with a subset of this Node's entries, filtered by query.

        Examples
        --------

        >>> from tiled.queries import FullText
        >>> await tree.search(FullText("hello"))
        """
        loop = asyncio.get_running_loop()
        client = await self.client
        return Catalog(await loop.run_in_executor(self.executor, client.search, query))

    async def distinct(
        self, *metadata_keys, structure_families=False, specs=False, counts=False
    ):
        """Get the unique values and optionally counts of metadata_keys,
        structure_families, and specs in this Node's entries

        Examples
        --------

        Query all the distinct values of a key.

        >>> await catalog.distinct("foo", counts=True)

        Query for multiple keys at once.

        >>> await catalog.distinct("foo", "bar", counts=True)

        """
        loop = asyncio.get_running_loop()
        client = await self.client
        query = partial(
            client.distinct,
            *metadata_keys,
            structure_families=structure_families,
            specs=specs,
            counts=counts,
        )
        return await loop.run_in_executor(self.executor, query)


# Create a default catalog for basic usage
catalog = Catalog()


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from ophyd_async.epics.adsimdetector import SimDetector as SimDetectorBase

from .area_detectors import HavenDetector


class SimDetector(HavenDetector, SimDetectorBase):
    _ophyd_labels_ = {"area_detectors", "detectors"}



import logging
from itertools import count
from typing import Mapping, Optional, Sequence

import numpy as np
from matplotlib.colors import TABLEAU_COLORS
from pandas.api.types import is_numeric_dtype
from pyqtgraph import GraphicsLayoutWidget, ImageView, PlotItem, PlotWidget
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import QFileDialog, QWidget

log = logging.getLogger(__name__)
colors = list(TABLEAU_COLORS.values())


class FiltersWidget(QWidget):
    returnPressed = Signal()

    def keyPressEvent(self, event):
        super().keyPressEvent(event)
        # Check for return keys pressed
        if event.key() in [Qt.Key_Enter, Qt.Key_Return]:
            self.returnPressed.emit()


class ExportDialog(QFileDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setFileMode(QFileDialog.FileMode.AnyFile)
        self.setAcceptMode(QFileDialog.AcceptSave)

    def ask(self, mimetypes: Optional[Sequence[str]] = None):
        """Get the name of the file to save for exporting."""
        self.setMimeTypeFilters(mimetypes)
        # Show the file dialog
        if self.exec_() == QFileDialog.Accepted:
            return self.selectedFiles()
        else:
            return None


class Browser1DPlotItem(PlotItem):
    hover_coords_changed = Signal(str)

    def hoverEvent(self, event):
        super().hoverEvent(event)
        if event.isExit():
            self.hover_coords_changed.emit("NaN")
            return
        # Get data coordinates from event
        pos = event.scenePos()
        data_pos = self.vb.mapSceneToView(pos)
        pos_str = f"({data_pos.x():.3f}, {data_pos.y():.3f})"
        self.hover_coords_changed.emit(pos_str)


class BrowserMultiPlotWidget(GraphicsLayoutWidget):
    _multiplot_items: Mapping

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._multiplot_items = {}

    def multiplot_items(self, n_cols: int = 3):
        view = self
        item0 = None
        for idx in count():
            row = int(idx / n_cols)
            col = idx % n_cols
            # Make a new plot item if one doesn't exist
            if (row, col) not in self._multiplot_items:
                self._multiplot_items[(row, col)] = view.addPlot(row=row, col=col)
            new_item = self._multiplot_items[(row, col)]
            # Link the X-axes together
            if item0 is None:
                item0 = new_item
            else:
                new_item.setXLink(item0)
            # Resize the viewing area to fit the contents
            width = view.width()
            plot_width = width / n_cols
            # view.resize(int(width), int(plot_width * row))
            view.setFixedHeight(1200)
            yield new_item

    def plot_runs(self, runs: Mapping, xsignal: str):
        """Take loaded run data and plot small multiples.

        Parameters
        ==========
        runs
          Dictionary with pandas series for each curve. The keys
          should be the curve labels, the series' indexes are the x
          values and the series' values are the y data.
        xsignal
          The name of the signal to use for the common horizontal
          axis.

        """
        # Use all the data columns as y signals
        ysignals = []
        for run in runs.values():
            ysignals.extend(run.columns)
        # Remove the x-signal from the list of y signals
        ysignals = sorted(list(dict.fromkeys(ysignals)))
        # Plot the runs
        self.clear()
        self._multiplot_items = {}
        for label, data in runs.items():
            # Figure out which signals to plot
            if xsignal in data.columns:
                xdata = data[xsignal].values
            else:
                # Could not find x signal, so use the index instead
                log.warning(
                    f"Cannot plot x='{xsignal}' for {list(data.keys())}. Falling back to index."
                )
                xdata = data.index
            # Plot each y signal on a separate plot
            for ysignal, plot_item in zip(ysignals, self.multiplot_items()):
                plot_item.setTitle(ysignal)
                if ysignal not in data.columns:
                    log.warning(f"No signal {ysignal} in data.")
                    continue
                ydata = data[ysignal].values
                if is_numeric_dtype(ydata):
                    plot_item.plot(xdata, ydata)
                log.debug(f"Plotted {ysignal} vs. {xsignal} for {data}")


class Browser1DPlotWidget(PlotWidget):
    auto_range_needed: bool
    data_items: dict

    def __init__(self, parent=None, background="default", plotItem=None, **kargs):
        plot_item = Browser1DPlotItem(**kargs)
        super().__init__(parent=parent, background=background, plotItem=plot_item)
        self.clear_runs()

    def clear_runs(self):
        self.getPlotItem().clear()
        self.cursor_line = None
        self.auto_range_needed = True
        self.data_items = {}

    def plot_runs(self, runs: Mapping, ylabel="", xlabel=""):
        """Take loaded run data and plot it.

        Parameters
        ==========
        runs
          Dictionary with pandas series for each curve. The keys
          should be the curve labels, the series' indexes are the x
          values and the series' values are the y data.

        """
        plot_item = self.getPlotItem()
        # Plot this run's data
        for idx, (label, series) in enumerate(runs.items()):
            color = colors[idx % len(colors)]
            if label in self.data_items.keys():
                # We've plotted this item before, so reuse it
                data_item = self.data_items[label]
                data_item.setData(series.index, series.values)
            else:
                self.data_items[label] = plot_item.plot(
                    x=series.index,
                    y=series.values,
                    pen=color,
                    name=label,
                    clear=False,
                )
            # Cursor to drag around on the data
            if self.cursor_line is None:
                self.cursor_line = plot_item.addLine(
                    x=np.median(series.index), movable=True, label="{value:.3f}"
                )
        # Axis formatting
        plot_item.setLabels(left=ylabel, bottom=xlabel)


class Browser2DPlotWidget(ImageView):
    """A plot widget for 2D maps."""

    def __init__(self, *args, view=None, **kwargs):
        if view is None:
            view = PlotItem()
        super().__init__(*args, view=view, **kwargs)

    def plot_runs(
        self, runs: Mapping, xlabel: str = "", ylabel: str = "", extents=None
    ):
        """Take loaded 2D or 3D mapping data and plot it.

        Parameters
        ==========
        runs
          Dictionary with pandas series for each curve. The keys
          should be the curve labels, the series' indexes are the x
          values and the series' values are the y data.
        xlabel
          The label for the horizontal axis.
        ylabel
          The label for the vertical axis.
        extents
          Spatial extents for the map as ((-y, +y), (-x, +x)).

        """
        images = np.asarray(list(runs.values()))
        # Combine the different runs into one image
        # To-do: make this respond to the combobox selection
        image = np.mean(images, axis=0)
        # To-do: Apply transformations

        # # Plot the image
        if 2 <= image.ndim <= 3:
            self.setImage(image.T, autoRange=False)
        else:
            log.info(f"Could not plot image of dataset with shape {image.shape}.")
            return
        # Determine the axes labels
        self.view.setLabel(axis="bottom", text=xlabel)
        self.view.setLabel(axis="left", text=ylabel)
        # Set axes extent
        yextent, xextent = extents
        x = xextent[0]
        y = yextent[0]
        w = xextent[1] - xextent[0]
        h = yextent[1] - yextent[0]
        self.getImageItem().setRect(x, y, w, h)



from ophyd_async.core import PathProvider, SubsetEnum
from ophyd_async.epics.adaravis import AravisDetector as DetectorBase
from ophyd_async.epics.core import epics_signal_rw_rbv

from .area_detectors import HavenDetector, default_path_provider


class AravisTriggerSource(SubsetEnum):
    SOFTWARE = "Software"
    LINE1 = "Line1"


class AravisDetector(HavenDetector, DetectorBase):
    _ophyd_labels_ = {"cameras", "detectors"}

    def __init__(
        self, prefix, *args, path_provider: PathProvider | None = None, **kwargs
    ):
        if path_provider is None:
            path_provider = default_path_provider()
        super().__init__(*args, prefix=prefix, path_provider=path_provider, **kwargs)
        # Replace a signal that has different enum options
        self.drv.trigger_source = epics_signal_rw_rbv(
            AravisTriggerSource,  # type: ignore
            f"{prefix}cam1:TriggerSource",
        )
        self.set_name(self.name)



import asyncio

import numpy as np
import pandas as pd
import pytest
from pyqtgraph import PlotItem

from firefly.xrf_detector import XRFDetectorDisplay

# detectors = ["dxp", "xspress"]
detectors = ["xspress"]


@pytest.fixture()
def xrf_display(request, qtbot):
    """Parameterized fixture for creating a display based on a specific
    detector class.

    """
    # Figure out which detector we're using
    det = request.getfixturevalue(request.param)
    # Create the display
    display = XRFDetectorDisplay(macros={"DEV": det.name})
    qtbot.addWidget(display)
    # Set sensible starting values
    spectra = np.random.default_rng(seed=0).integers(
        0, 65536, dtype=np.int_, size=(4, 1024)
    )
    plot_widget = display.mca_plot_widget
    energies = np.arange(1024)
    plot_widget.update_spectrum(0, pd.Series(spectra[0], index=energies))
    plot_widget.update_spectrum(1, pd.Series(spectra[1], index=energies))
    plot_widget.update_spectrum(2, pd.Series(spectra[2], index=energies))
    plot_widget.update_spectrum(3, pd.Series(spectra[3], index=energies))
    yield display


@pytest.mark.parametrize("xrf_display", detectors, indirect=True)
def test_device_name(xrf_display):
    label = xrf_display.ui.detector_name_label
    assert label.text() == xrf_display.device.name


@pytest.mark.parametrize("xrf_display", detectors, indirect=True)
def test_mca_count_labels_created(xrf_display):
    """Check that QLabel objs are created for each element."""
    layout = xrf_display.ui.mcas_layout
    assert layout.rowCount() == 6  # 4 elements plus heading and total
    assert layout.itemAtPosition(1, 0).widget().text() == "Total"
    assert layout.itemAtPosition(5, 1).widget() is xrf_display._count_labels[3]


@pytest.mark.parametrize("xrf_display", detectors, indirect=True)
async def test_update_mca_spectra(xrf_display, qtbot):
    spectra = np.random.default_rng(seed=0).integers(
        0, 65536, dtype=np.int_, size=(4, 1024)
    )
    mca_plot_widget = xrf_display.ui.mca_plot_widget
    # Check that a PlotItem was created in the fixture
    plot_item = mca_plot_widget.ui.plot_widget.getPlotItem()
    assert isinstance(plot_item, PlotItem)
    # Clear the data items so we can test them later
    plot_item.clear()
    # plot_widget.update_spectrum(spectrum=spectra[0], mca_idx=1)
    with qtbot.waitSignal(mca_plot_widget.plot_changed):
        xrf_display._spectrum_channels[0].value_slot(spectra[0])
        xrf_display._spectrum_channels[1].value_slot(spectra[1])
        await asyncio.sleep(0.1)
    # Check that the spectrum was plotted
    data_items = plot_item.listDataItems()
    assert len(data_items) == 2
    energies = np.linspace(5, 10235, num=1024)
    np.testing.assert_equal(data_items[0].xData, energies)
    # Check that previous plots get cleared
    spectra2 = np.random.default_rng(seed=1).integers(
        0, 65536, dtype=np.int_, size=(4, 1024)
    )
    with qtbot.waitSignal(mca_plot_widget.plot_changed):
        xrf_display._spectrum_channels[0].value_slot(spectra2[0])
        await asyncio.sleep(0.1)
    data_items = plot_item.listDataItems()
    assert len(data_items) == 2


@pytest.mark.xfail()
@pytest.mark.parametrize("xrf_display", detectors, indirect=True)
def test_mca_hovering(xrf_display):
    """Is the spectrum highlighted when the element row is selected."""
    # Check that the region is hidden at startup
    plot_widget = xrf_display.mca_plot_widget
    # Now highlight a spectrum, and confirm it is visible
    plot_widget.highlight_spectrum(mca_num=2, roi_num=0, hovered=True)
    # assert region.brush.color().name() == "#ff7f0e"
    # Unhighlight and confirm it is invisible
    plot_widget.highlight_spectrum(mca_num=1, roi_num=0, hovered=False)


@pytest.mark.parametrize("xrf_display", detectors, indirect=True)
def test_update_spectral_widgets(xrf_display):
    spectrum = np.ones(shape=(100,), dtype=int) * 10
    xrf_display.update_spectral_widgets(mca_num=0, spectrum=spectrum, spectra=[])
    mcas_layout = xrf_display.ui.mcas_layout
    elem0_label = mcas_layout.itemAtPosition(2, 1).widget()
    assert elem0_label.text() == "1_000"
    total_label = mcas_layout.itemAtPosition(1, 1).widget()
    assert total_label.text() == "0"
    # Add a second spectrum for a separate element
    xrf_display.update_spectral_widgets(
        mca_num=1, spectrum=spectrum * 2, spectra=[spectrum, spectrum * 2]
    )
    elem1_label = mcas_layout.itemAtPosition(3, 1).widget()
    assert elem1_label.text() == "2_000"
    total_label = mcas_layout.itemAtPosition(1, 1).widget()
    assert total_label.text() == "3_000"


@pytest.mark.parametrize("xrf_display", detectors, indirect=True)
def test_detector_state_style(xrf_display):
    """The label should change color, etc depending on detector state."""
    # Check initial state
    lbl = xrf_display.ui.detector_state_label
    assert "rgb(" not in lbl.styleSheet()
    assert "bold" not in lbl.styleSheet()
    # Update state and check again
    xrf_display.update_state_style("Acquire")
    assert "rgb(" in lbl.styleSheet()
    assert "bold" in lbl.styleSheet()


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------
